# -*- coding: utf-8 -*-
"""Module to handle the importation from MRIFileManager and its progress

:Contains:
    :Class:
        - ImportProgress : Inherit from QProgressDialog and handle the
          progress bar
        - ImportWorker : Inherit from QThread and manage the threads

    :Methods:
        - read_log : Show the evolution of the progress bar and returns its
          feedback
        - tags_from_file : Returns a list of [tag, value] contained in a Json
          file
        - verify_scans : Check if the project's scans have been modified

"""

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

# import datetime
import glob
import hashlib  # To generate the md5 of each path
import json
import os.path
import threading
import traceback
from datetime import datetime
from time import sleep, time

# Populse_db imports
from populse_db.database import (
    FIELD_TYPE_BOOLEAN,
    FIELD_TYPE_DATE,
    FIELD_TYPE_DATETIME,
    FIELD_TYPE_FLOAT,
    FIELD_TYPE_INTEGER,
    FIELD_TYPE_LIST_BOOLEAN,
    FIELD_TYPE_LIST_DATE,
    FIELD_TYPE_LIST_DATETIME,
    FIELD_TYPE_LIST_FLOAT,
    FIELD_TYPE_LIST_INTEGER,
    FIELD_TYPE_LIST_STRING,
    FIELD_TYPE_LIST_TIME,
    FIELD_TYPE_STRING,
    FIELD_TYPE_TIME,
)

# PyQt5 imports
from PyQt5.QtCore import Qt, QThread, pyqtSignal
from PyQt5.QtWidgets import QProgressDialog

from populse_mia.data_manager.database_mia import (
    TAG_ORIGIN_BUILTIN,
    TAG_ORIGIN_USER,
)

# Populse_MIA imports
from populse_mia.data_manager.project import (
    COLLECTION_CURRENT,
    COLLECTION_INITIAL,
    TAG_CHECKSUM,
    TAG_FILENAME,
    TAG_TYPE,
    TYPE_BVAL,
    TYPE_BVEC,
    TYPE_BVEC_BVAL,
    TYPE_NII,
)


class ImportProgress(QProgressDialog):
    """Handle the progress bar.

    :param project: A Project object

    .. Methods:
        - onProgress : Set the import progressbar value.

    """

    def __init__(self, project):
        super(ImportProgress, self).__init__(
            "Please wait while the paths are being imported...", None, 0, 3
        )

        self.setWindowTitle("Importing the paths")
        self.setWindowFlags(
            Qt.Window | Qt.WindowTitleHint | Qt.CustomizeWindowHint
        )
        self.setModal(True)

        self.setMinimumDuration(0)
        self.setValue(0)
        self.setMinimumWidth(350)  # For mac OS

        self.worker = ImportWorker(project, self)
        self.worker.finished.connect(self.close)
        self.worker.notifyProgress.connect(self.onProgress)
        self.worker.start()

    def onProgress(self, i):
        """Signal to set the import progressbar value

        :param i: int, value of the progressbar
        """

        self.setValue(i)


class ImportWorker(QThread):
    """Manage threads.

    :param project: A Project object
    :param progress: An ImportProgress object

    .. Methods:
        - run : Override the QThread run method.
    """

    # Used to fill the progress bar
    notifyProgress = pyqtSignal(int)

    def __init__(self, project, progress):
        super().__init__()
        self.project = project
        self.progress = progress
        self.lock = threading.RLock()
        # scans_added should always be accessed through the lock, and copied
        # before releasing the lock, because its value will change inside
        # the thread
        self.scans_added = []

    def run(self):
        """Override the QThread run method. Executed when the worker is
        started, fills the database and updates the progress.
        """
        begin = time()

        raw_data_folder = os.path.relpath(
            os.path.join(self.project.folder, "data", "raw_data")
        )

        # Checking all the export logs from MRIManager and taking the most
        # recent
        list_logs = glob.glob(os.path.join(raw_data_folder, "logExport*.json"))

        if len(list_logs) == 0:
            list_dict_log = []

        else:
            log_to_read = max(list_logs, key=os.path.getctime)

            with open(log_to_read, "r", encoding="utf-8") as file:
                list_dict_log = json.load(file)

        # For history
        historyMaker = list()
        historyMaker.append("add_scans")
        with self.lock:
            self.scans_added = []
        values_added = []
        tags_added = []
        tags_names_added = []
        documents = {}

        # List of tags to remove
        tags_to_remove = ["Dataset data file", "Dataset header file"]

        for dict_log in list_dict_log:
            if dict_log["StatusExport"] == "Export ok":
                file_name = dict_log["NameFile"]
                path_name = raw_data_folder

                with open(
                    os.path.join(path_name, file_name) + ".nii", "rb"
                ) as scan_file:
                    data = scan_file.read()
                    original_md5 = hashlib.md5(data).hexdigest()

                file_path = os.path.join(raw_data_folder, file_name + ".nii")
                file_database_path = os.path.relpath(
                    file_path, self.project.folder
                )

                document_not_existing = (
                    self.project.session.get_document(
                        COLLECTION_CURRENT, file_database_path
                    )
                    is None
                )

                if document_not_existing:
                    with self.lock:
                        # Scan added to history
                        self.scans_added.append(file_database_path)

                documents[file_database_path] = {}
                documents[file_database_path][
                    TAG_FILENAME
                ] = file_database_path

                # print('\ntags_from_file(file_name, path_name): ',
                # tags_from_file(file_name, path_name))

                # For each tag in each scan
                for tag in tags_from_file(file_name, path_name):
                    # We do the tag only if it's not in the tags to remove
                    if tag[0] not in tags_to_remove:
                        tag_name = tag[0]

                        # print('\ntag_name: ', tag_name)

                        properties = tag[1]

                        # print('properties: ', properties)

                        format = ""
                        description = None
                        unit = None
                        tag_type = FIELD_TYPE_STRING

                        if isinstance(properties, dict):
                            format = properties["format"]

                            if properties["description"] != "":
                                description = properties["description"]

                            if properties["units"] != "":
                                unit = properties["units"]

                            if properties["type"] != "":
                                tag_type = properties["type"]

                            value = properties["value"]

                        else:
                            if isinstance(properties, list):
                                value = properties[0]
                            else:
                                value = properties

                        # Creating date types
                        if format is not None and format != "":
                            format = format.replace("yyyy", "%Y")
                            format = format.replace("MM", "%m")
                            format = format.replace("dd", "%d")
                            format = format.replace("HH", "%H")
                            format = format.replace("mm", "%M")
                            format = format.replace("ss", "%S")
                            format = format.replace("SSS", "%f")

                            if (
                                "%Y" in format
                                and "%m" in format
                                and "%d" in format
                                and "%H" in format
                                and "%M" in format
                                and "%S" in format
                            ):
                                tag_type = FIELD_TYPE_DATETIME

                            elif (
                                "%Y" in format
                                and "%m" in format
                                and "%d" in format
                            ):
                                tag_type = FIELD_TYPE_DATE

                            elif (
                                "%H" in format
                                and "%M" in format
                                and "%S" in format
                            ):
                                tag_type = FIELD_TYPE_TIME

                        if tag_name != "Json_Version":
                            # Preparing value and type
                            if hasattr(value, "__len__") and not isinstance(
                                value, str
                            ):
                                if (
                                    len(value) == 1
                                    and isinstance(value[0], list)
                                ) or (len(value) != 1):
                                    if tag_type == FIELD_TYPE_STRING:
                                        tag_type = FIELD_TYPE_LIST_STRING

                                    elif tag_type == FIELD_TYPE_INTEGER:
                                        tag_type = FIELD_TYPE_LIST_INTEGER

                                    elif tag_type == FIELD_TYPE_FLOAT:
                                        tag_type = FIELD_TYPE_LIST_FLOAT

                                    elif tag_type == FIELD_TYPE_BOOLEAN:
                                        tag_type = FIELD_TYPE_LIST_BOOLEAN

                                    elif tag_type == FIELD_TYPE_DATE:
                                        tag_type = FIELD_TYPE_LIST_DATE

                                    elif tag_type == FIELD_TYPE_DATETIME:
                                        tag_type = FIELD_TYPE_LIST_DATETIME

                                    elif tag_type == FIELD_TYPE_TIME:
                                        tag_type = FIELD_TYPE_LIST_TIME

                                if len(value) == 1:
                                    value = value[0]

                                else:
                                    value_prepared = []

                                    for value_single in value:
                                        value_prepared.append(value_single[0])

                                    value = value_prepared
                        # print('value: ', value)
                        # print('tag_type: ', tag_type)

                        if (
                            tag_type == FIELD_TYPE_DATETIME
                            or tag_type == FIELD_TYPE_DATE
                            or tag_type == FIELD_TYPE_TIME
                        ):
                            if value is not None and value != "":
                                value = datetime.strptime(value, format)

                                if tag_type == FIELD_TYPE_TIME:
                                    value = value.time()

                                elif tag_type == FIELD_TYPE_DATE:
                                    value = value.date()

                        # TODO time lists

                        tag_row = self.project.session.get_field(
                            COLLECTION_CURRENT, tag_name
                        )

                        if (
                            tag_row is None
                            and tag_name not in tags_names_added
                        ):
                            # Adding the tag as it's not in the database yet
                            tags_added.append(
                                [
                                    COLLECTION_CURRENT,
                                    tag_name,
                                    tag_type,
                                    description,
                                    False,
                                    TAG_ORIGIN_BUILTIN,
                                    unit,
                                    None,
                                ]
                            )
                            tags_added.append(
                                [
                                    COLLECTION_INITIAL,
                                    tag_name,
                                    tag_type,
                                    description,
                                    False,
                                    TAG_ORIGIN_BUILTIN,
                                    unit,
                                    None,
                                ]
                            )
                            tags_names_added.append(tag_name)

                        # The value is accepted if it's not empty or null
                        if value is not None and value != "":
                            if document_not_existing:
                                values_added.append(
                                    [
                                        file_database_path,
                                        tag_name,
                                        value,
                                        value,
                                    ]
                                )  # Value added to history
                            documents[file_database_path][tag_name] = value

                if document_not_existing:
                    # Tags added manually
                    # Value added to history
                    values_added.append(
                        [
                            file_database_path,
                            TAG_CHECKSUM,
                            original_md5,
                            original_md5,
                        ]
                    )
                    # Value added to history
                    values_added.append(
                        [file_database_path, TAG_TYPE, TYPE_NII, TYPE_NII]
                    )
                documents[file_database_path][TAG_CHECKSUM] = original_md5
                documents[file_database_path][TAG_TYPE] = TYPE_NII

                # Index bvec/ bval in the database if they exist
                if dict_log["Bvec_bval"] == "yes":
                    file_name = dict_log["NameFile"]
                    # bvec /bval FSL format
                    bvec_path = os.path.join(
                        raw_data_folder, file_name + ".bvec"
                    )
                    bval_path = os.path.join(
                        raw_data_folder, file_name + ".bval"
                    )
                    # bvec / bval mrtrix format
                    bvec_bval_mrtrix_path = os.path.join(
                        raw_data_folder, file_name + "-bvecs-bvals-MRtrix.txt"
                    )

                    if (
                        os.path.exists(bvec_path) and os.path.exists(bval_path)
                    ) or os.path.exists(bvec_bval_mrtrix_path):
                        # Add a tag to link bvec/bval and the nifti file
                        tag_name = "AssociatedNIfTIFile"
                        tag_type = FIELD_TYPE_STRING
                        description = "Associated NIfTI file"
                        unit = None

                        tag_row = self.project.session.get_field(
                            COLLECTION_CURRENT, tag_name
                        )

                        if (
                            tag_row is None
                            and tag_name not in tags_names_added
                        ):
                            # Adding the tag as it's not in the database yet
                            tags_added.append(
                                [
                                    COLLECTION_CURRENT,
                                    tag_name,
                                    tag_type,
                                    description,
                                    False,
                                    TAG_ORIGIN_BUILTIN,
                                    unit,
                                    None,
                                ]
                            )
                            tags_added.append(
                                [
                                    COLLECTION_INITIAL,
                                    tag_name,
                                    tag_type,
                                    description,
                                    False,
                                    TAG_ORIGIN_BUILTIN,
                                    unit,
                                    None,
                                ]
                            )
                            tags_names_added.append(tag_name)

                    if os.path.exists(bvec_path) and os.path.exists(bval_path):
                        # Index BVEC FSL format in the database
                        with open(bvec_path, "rb") as bvec_file:
                            original_md5_bvec = hashlib.md5(
                                bvec_file.read()
                            ).hexdigest()

                        bvec_database_path = os.path.relpath(
                            bvec_path, self.project.folder
                        )

                        document_not_existing = (
                            self.project.session.get_document(
                                COLLECTION_CURRENT, bvec_database_path
                            )
                            is None
                        )

                        if document_not_existing:
                            with self.lock:
                                # Scan added to history
                                self.scans_added.append(bvec_database_path)
                        documents[bvec_database_path] = {}
                        documents[bvec_database_path][
                            TAG_FILENAME
                        ] = bvec_database_path

                        if document_not_existing:
                            # Tags added manually
                            # Value added to history
                            values_added.append(
                                [
                                    bvec_database_path,
                                    TAG_CHECKSUM,
                                    original_md5_bvec,
                                    original_md5_bvec,
                                ]
                            )
                            # Value added to history
                            values_added.append(
                                [
                                    bvec_database_path,
                                    TAG_TYPE,
                                    TYPE_BVEC,
                                    TYPE_BVEC,
                                ]
                            )

                            # Value added to history
                            values_added.append(
                                [
                                    bvec_database_path,
                                    tag_name,
                                    file_database_path,
                                    file_database_path,
                                ]
                            )

                        documents[bvec_database_path][
                            TAG_CHECKSUM
                        ] = original_md5_bvec
                        documents[bvec_database_path][TAG_TYPE] = TYPE_BVEC
                        documents[bvec_database_path][tag_name] = str(
                            file_database_path
                        )

                        # Index BVAL FSL format in the database
                        with open(bval_path, "rb") as bval_file:
                            original_md5_bval = hashlib.md5(
                                bval_file.read()
                            ).hexdigest()
                        bval_database_path = os.path.relpath(
                            bval_path, self.project.folder
                        )

                        document_not_existing = (
                            self.project.session.get_document(
                                COLLECTION_CURRENT, bval_database_path
                            )
                            is None
                        )

                        if document_not_existing:
                            with self.lock:
                                # Scan added to history
                                self.scans_added.append(bval_database_path)
                        documents[bval_database_path] = {}
                        documents[bval_database_path][
                            TAG_FILENAME
                        ] = bval_database_path

                        if document_not_existing:
                            # Tags added manually
                            # Value added to history
                            values_added.append(
                                [
                                    bval_database_path,
                                    TAG_CHECKSUM,
                                    original_md5_bval,
                                    original_md5_bval,
                                ]
                            )
                            # Value added to history
                            values_added.append(
                                [
                                    bval_database_path,
                                    TAG_TYPE,
                                    TYPE_BVAL,
                                    TYPE_BVAL,
                                ]
                            )
                            # Value added to history
                            values_added.append(
                                [
                                    bval_database_path,
                                    tag_name,
                                    file_database_path,
                                    file_database_path,
                                ]
                            )

                        documents[bval_database_path][
                            TAG_CHECKSUM
                        ] = original_md5_bval
                        documents[bval_database_path][TAG_TYPE] = TYPE_BVAL
                        documents[bval_database_path][tag_name] = str(
                            file_database_path
                        )
                    if os.path.exists(bvec_bval_mrtrix_path):
                        # Index BVEC/BVAL MRTRIX format in the database
                        with open(
                            bvec_bval_mrtrix_path, "rb"
                        ) as bval_bvec_file:
                            original_md5_bvec_bval = hashlib.md5(
                                bval_bvec_file.read()
                            ).hexdigest()

                        bvec_bval_mrtrix_database_path = os.path.relpath(
                            bvec_bval_mrtrix_path, self.project.folder
                        )

                        document_not_existing = (
                            self.project.session.get_document(
                                COLLECTION_CURRENT,
                                bvec_bval_mrtrix_database_path,
                            )
                            is None
                        )

                        if document_not_existing:
                            with self.lock:
                                # Scan added to history
                                self.scans_added.append(
                                    bvec_bval_mrtrix_database_path
                                )
                        documents[bvec_bval_mrtrix_database_path] = {}
                        documents[bvec_bval_mrtrix_database_path][
                            TAG_FILENAME
                        ] = bvec_bval_mrtrix_database_path

                        if document_not_existing:
                            # Tags added manually
                            # Value added to history
                            values_added.append(
                                [
                                    bvec_bval_mrtrix_database_path,
                                    TAG_CHECKSUM,
                                    original_md5_bvec_bval,
                                    original_md5_bvec_bval,
                                ]
                            )
                            # Value added to history
                            values_added.append(
                                [
                                    bvec_bval_mrtrix_database_path,
                                    TAG_TYPE,
                                    TYPE_BVEC_BVAL,
                                    TYPE_BVEC_BVAL,
                                ]
                            )

                            # Value added to history
                            values_added.append(
                                [
                                    bvec_bval_mrtrix_database_path,
                                    tag_name,
                                    file_database_path,
                                    file_database_path,
                                ]
                            )

                        documents[bvec_bval_mrtrix_database_path][
                            TAG_CHECKSUM
                        ] = original_md5_bvec_bval
                        documents[bvec_bval_mrtrix_database_path][
                            TAG_TYPE
                        ] = TYPE_BVEC_BVAL
                        documents[bvec_bval_mrtrix_database_path][tag_name] = (
                            str(file_database_path)
                        )

        # Missing values added thanks to default values
        for tag in self.project.session.get_fields(COLLECTION_CURRENT):
            if tag.origin == TAG_ORIGIN_USER:
                for scan in self.scans_added:
                    if (
                        tag.default_value is not None
                        and self.project.session.get_value(
                            COLLECTION_CURRENT, scan[0], tag.field_name
                        )
                        is None
                    ):
                        # Value added to history
                        values_added.append(
                            [
                                scan,
                                tag.field_name,
                                tag.default_value,
                                tag.default_value,
                            ]
                        )
                        documents[scan][tag.field_name] = tag.default_value

        self.notifyProgress.emit(1)
        sleep(0.1)

        self.project.session.add_fields(tags_added)

        self.notifyProgress.emit(2)
        sleep(0.1)

        current_paths = self.project.session.get_documents_names(
            COLLECTION_CURRENT
        )

        for document in documents:
            if document in current_paths:
                self.project.session.remove_document(
                    COLLECTION_CURRENT, document
                )
                self.project.session.remove_document(
                    COLLECTION_INITIAL, document
                )

            self.project.session.add_document(
                COLLECTION_CURRENT, documents[document], flush=False
            )
            self.project.session.add_document(
                COLLECTION_INITIAL, documents[document], flush=False
            )

        self.project.session.commit()
        self.notifyProgress.emit(3)
        sleep(0.1)

        # For history
        historyMaker.append(self.scans_added)
        historyMaker.append(values_added)
        self.project.undos.append(historyMaker)
        self.project.redos.clear()

        print("\nData export duration in the database:")
        print("read_log time: " + str(round(time() - begin, 2)) + " s\n")
        # print("read_log time: " + str(time() - begin))

        # pr.disable()
        # pr.print_stats(sort='time')
        # prof.print_stats()


def read_log(project, main_window):
    """Show the evolution of the progress bar and returns its feedback, a list
    of the paths to each data file that was loaded.

    :param project: current project in the software
    :param main_window: software's main window
    :returns: the scans that have been added
    """

    main_window.progress = ImportProgress(project)
    main_window.progress.show()
    main_window.progress.exec()

    with main_window.progress.worker.lock:
        scans_added = list(main_window.progress.worker.scans_added)
    return scans_added


# def save_project(project):
#     """
#     Saves the modifications of the project
#
#     :param project: current project in the software
#     """
#
#     project.saveModifications()


def tags_from_file(file_path, path):
    """Return a list of [tag, value] contained in a Json file.

    :param file_path: file path of the Json file (without the extension)
    :param path: project path
    :returns: a list of the Json tags of the file
    """
    json_tags = []

    with open(os.path.join(path, file_path) + ".json") as f:

        for name, value in json.load(f).items():

            # We don't want spaces in PatientName (used by Mia to define
            # subfolders when writing calculation results)
            if name == "PatientName":
                value["value"] = [value["value"][0].replace(" ", "")]

            json_tags.append([name, value])

    return json_tags


def verify_scans(project):
    """Check if the project's scans have been modified.

    :param project: current project in the software
    :returns: the list of scans that have been modified
    """

    # Returning the files that are problematic
    return_list = []
    for scan in project.session.get_documents_names(COLLECTION_CURRENT):
        file_name = scan
        file_path = os.path.relpath(os.path.join(project.folder, file_name))

        if os.path.exists(file_path):
            # If the file exists, we do the checksum
            with open(file_path, "rb") as scan_file:
                try:
                    data = scan_file.read()

                except Exception as e:
                    print(
                        '\nError while reading the "{0}" file '
                        "...!\nTraceback:".format(os.path.abspath(file_path))
                    )
                    print(
                        "".join(traceback.format_tb(e.__traceback__)), end=""
                    )
                    print("{0}: {1}\n".format(e.__class__.__name__, e))
                    actual_md5 = None

                else:
                    actual_md5 = hashlib.md5(data).hexdigest()

            initial_checksum = project.session.get_value(
                COLLECTION_CURRENT, scan, TAG_CHECKSUM
            )
            if initial_checksum is not None and actual_md5 != initial_checksum:
                return_list.append(file_name)

        else:
            # Otherwise, we directly add the file in the list
            return_list.append(file_name)

    return return_list



# -*- coding: utf-8 -*-
"""Module that handles pipeline iteration.

:Contains:
    :Class:
        - IterationTable

"""

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################


import os

# PyQt5 imports
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import (
    QCheckBox,
    QComboBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
    QWidget,
)

from populse_mia.data_manager.project import COLLECTION_CURRENT, TAG_FILENAME
from populse_mia.software_properties import Config

# MIA imports
from populse_mia.user_interface.pipeline_manager.process_mia import ProcessMIA
from populse_mia.user_interface.pop_ups import (
    ClickableLabel,
    PopUpSelectIteration,
    PopUpSelectTagCountTable,
)


class IterationTable(QWidget):
    """Widget that handles pipeline iteration.

    .. Methods:
        - add_tag: adds a tag to visualize in the iteration table
        - emit_iteration_table_updated: emits a signal when the iteration
                                        scans have been updated
        - fill_values: fill values_list depending on the visualized tags
        - filter_values: select the tag values used for the iteration
        - refresh_layout: updates the layout of the widget
        - remove_tag: removes a tag to visualize in the iteration table
        - select_iterated_tag: opens a pop-up to let the user select on which
                               tag to iterate
        - select_visualized_tag: opens a pop-up to let the user select which
                                 tag to visualize in the iteration table
        - update_iterated_tag: updates the widget
        - update_table: updates the iteration table
        - update_selected_tag: updates the selected tag for current pipeline
                               manager tab

    """

    iteration_table_updated = pyqtSignal(list, list)

    def __init__(self, project, scan_list, main_window):
        """
        Initialization of the IterationTable widget.

        :param project: current project in the software
        :param scan_list: list of the selected database files
        :param main_window: software's main_window
        """

        QWidget.__init__(self)

        # Necessary for using MIA bricks
        ProcessMIA.project = project

        self.project = project

        if not scan_list:
            self.scan_list = self.project.session.get_documents_names(
                COLLECTION_CURRENT
            )
        else:
            self.scan_list = scan_list

        self.main_window = main_window

        # values_list will contain the different values of each selected tag
        self.values_list = [[], []]
        self.all_tag_values = []

        # Checkbox to choose to iterate the pipeline or not
        self.check_box_iterate = QCheckBox("Iterate pipeline")
        self.check_box_iterate.stateChanged.connect(
            self.emit_iteration_table_updated
        )

        # Label "Iterate over:"
        self.label_iterate = QLabel("Iterate over:")

        # Label that displays the name of the selected tag
        self.iterated_tag_label = QLabel("Select a tag")

        # Push button to select the tag to iterate
        self.iterated_tag_push_button = QPushButton("Select")
        self.iterated_tag_push_button.clicked.connect(
            self.select_iteration_tag
        )

        # QComboBox
        self.combo_box = QComboBox()
        self.combo_box.currentIndexChanged.connect(self.update_table)

        # filter
        self.filter_button = QPushButton("Filter")
        self.filter_button.clicked.connect(self.filter_values)

        # QTableWidget
        self.iteration_table = QTableWidget()

        # Label tag
        self.label_tags = QLabel("Tags to visualize:")

        # Each push button will allow the user to visualize a tag in
        # the iteration browser
        push_button_tag_1 = QPushButton()
        push_button_tag_1.setText("SequenceName")
        push_button_tag_1.clicked.connect(
            lambda: self.select_visualized_tag(0)
        )

        push_button_tag_2 = QPushButton()
        push_button_tag_2.setText("AcquisitionDate")
        push_button_tag_2.clicked.connect(
            lambda: self.select_visualized_tag(1)
        )

        # The list of all the push buttons
        # (the user can add as many as he or she wants)
        self.push_buttons = []
        self.push_buttons.insert(0, push_button_tag_1)
        self.push_buttons.insert(1, push_button_tag_2)

        # Labels to add/remove a tag (a push button)
        self.add_tag_label = ClickableLabel()
        self.add_tag_label.setObjectName("plus")
        sources_images_dir = Config().getSourceImageDir()
        add_tag_picture = QPixmap(
            os.path.relpath(os.path.join(sources_images_dir, "green_plus.png"))
        )
        add_tag_picture = add_tag_picture.scaledToHeight(15)
        self.add_tag_label.setPixmap(add_tag_picture)
        self.add_tag_label.clicked.connect(self.add_tag)

        self.remove_tag_label = ClickableLabel()
        remove_tag_picture = QPixmap(
            os.path.relpath(os.path.join(sources_images_dir, "red_minus.png"))
        )
        remove_tag_picture = remove_tag_picture.scaledToHeight(20)
        self.remove_tag_label.setPixmap(remove_tag_picture)
        self.remove_tag_label.clicked.connect(self.remove_tag)

        # Layout
        self.v_layout = QVBoxLayout()
        self.setLayout(self.v_layout)
        self.refresh_layout()

    def add_tag(self):
        """Add a tag to visualize in the iteration table.

        Used only for tests
        """

        idx = len(self.push_buttons)
        push_button = QPushButton()
        push_button.setText("Tag n°" + str(len(self.push_buttons) + 1))
        push_button.clicked.connect(lambda: self.select_visualized_tag(idx))
        self.push_buttons.insert(len(self.push_buttons), push_button)
        self.refresh_layout()

    def emit_iteration_table_updated(self):
        """Emit a signal when the iteration scans have been updated."""
        if self.check_box_iterate.checkState():
            if hasattr(self, "scans"):
                self.iteration_table_updated.emit(
                    self.iteration_scans, self.all_iterations_scans
                )
            else:
                self.iteration_table_updated.emit(
                    self.scan_list, [self.scan_list]
                )
        else:
            self.iteration_table_updated.emit(self.scan_list, [self.scan_list])

    def fill_values(self, idx):
        """
        Fill values_list depending on the visualized tags

        :param idx: Index of the tag
        """
        tag_name = self.push_buttons[idx].text()
        values = []
        for scan in self.project.session.get_documents_names(
            COLLECTION_CURRENT
        ):
            current_value = self.project.session.get_value(
                COLLECTION_CURRENT, scan, tag_name
            )
            if current_value is not None:
                values.append(current_value)

        idx_to_fill = len(self.values_list)
        while len(self.values_list) <= idx:
            self.values_list.insert(idx_to_fill, [])
            idx_to_fill += 1

        if self.values_list[idx] is not None:
            self.values_list[idx] = []

        for value in values:
            if value not in self.values_list[idx]:
                self.values_list[idx].append(value)

    def filter_values(self):
        """Select the tag values used for the iteration"""

        # fmt: off
        iterated_tag = (
            self.main_window.pipeline_manager.pipelineEditorTabs.
            get_current_editor().iterated_tag
        )
        tag_values = (
            self.main_window.pipeline_manager.pipelineEditorTabs.
            get_current_editor().all_tag_values_list
        )
        # fmt: on

        ui_iteration = PopUpSelectIteration(iterated_tag, tag_values)

        if ui_iteration.exec_():
            tag_values_list = [
                t.replace("&", "") for t in ui_iteration.final_values
            ]
            # fmt: off
            (
                self.main_window.pipeline_manager.pipelineEditorTabs.
                get_current_editor
            )().tag_values_list = tag_values_list
            # fmt: on

            self.combo_box.clear()
            self.combo_box.addItems(tag_values_list)
            self.update_table()

    def refresh_layout(self):
        """Update the layout of the widget.

        Called in widget's initialization and when a tag push button
        is added or removed.
        """

        first_v_layout = QVBoxLayout()
        first_v_layout.addWidget(self.check_box_iterate)

        second_v_layout = QVBoxLayout()
        second_v_layout.addWidget(self.label_iterate)
        second_v_layout.addWidget(self.iterated_tag_label)

        third_v_layout = QVBoxLayout()
        third_v_layout.addWidget(self.iterated_tag_push_button)
        hbox = QHBoxLayout()
        hbox.addWidget(self.combo_box)
        hbox.addWidget(self.filter_button)
        third_v_layout.addLayout(hbox)

        top_layout = QHBoxLayout()
        top_layout.addLayout(first_v_layout)
        top_layout.addLayout(second_v_layout)
        top_layout.addLayout(third_v_layout)

        self.v_layout.addLayout(top_layout)
        self.v_layout.addWidget(self.iteration_table)

        self.h_box = QHBoxLayout()
        self.h_box.setSpacing(10)
        self.h_box.addWidget(self.label_tags)

        for tag_label in self.push_buttons:
            self.h_box.addWidget(tag_label)

        self.h_box.addWidget(self.add_tag_label)
        self.h_box.addWidget(self.remove_tag_label)
        self.h_box.addStretch(1)

        self.v_layout.addLayout(self.h_box)

    def remove_tag(self):
        """Remove a tag to visualize in the iteration table."""
        if len(self.push_buttons) >= 1:
            push_button = self.push_buttons[-1]
            push_button.deleteLater()
            push_button = None
            del self.push_buttons[-1]
            del self.values_list[-1]
            self.refresh_layout()
            self.update_table()

    def select_iteration_tag(self):
        """Open a pop-up to let the user select on which tag to iterate."""

        # fmt: off
        ui_select = PopUpSelectTagCountTable(
            self.project,
            self.project.session.get_fields_names(COLLECTION_CURRENT),
            self.main_window.pipeline_manager.pipelineEditorTabs.
            get_current_editor().iterated_tag,
        )
        # fmt: on

        if ui_select.exec_():
            # fmt: off
            if (
                self.main_window.pipeline_manager.pipelineEditorTabs.
                get_current_editor().iterated_tag is None
                and ui_select.selected_tag is None
            ):
                # fmt: on
                pass

            else:
                (self.main_window.pipeline_manager.pipelineEditorTabs.
                 get_current_editor)().iterated_tag = ui_select.selected_tag

                # Retrieve tag values
                self.update_selected_tag(ui_select.selected_tag)

    def select_visualized_tag(self, idx):
        """Open a pop-up to let the user select which tag to visualize in the
        iteration table.

        :param idx: index of the clicked push button
        """

        popUp = PopUpSelectTagCountTable(
            self.project,
            self.project.session.get_fields_names(COLLECTION_CURRENT),
            self.push_buttons[idx].text(),
        )
        if popUp.exec_() and popUp.selected_tag is not None:
            self.push_buttons[idx].setText(popUp.selected_tag)
            self.fill_values(idx)
            self.update_table()

    def update_iterated_tag(self, tag_name=None):
        """
        Update the widget when the iterated tag is modified.

        :param tag_name: name of the iterated tag
        """

        if len(self.main_window.pipeline_manager.scan_list) > 0:
            self.scan_list = self.main_window.pipeline_manager.scan_list
        else:
            self.scan_list = self.project.session.get_documents_names(
                COLLECTION_CURRENT
            )

        self.combo_box.clear()
        if tag_name is None:
            self.iterated_tag_push_button.setText("Select")
            self.iterated_tag_label.setText("Select a tag")
            self.iteration_table.clear()
            self.iteration_table.setColumnCount(len(self.push_buttons))
        else:
            self.iterated_tag_push_button.setText(tag_name)
            self.iterated_tag_label.setText(tag_name + ":")

            # duplicate the values list to have the initial, unfiltered, one
            # fmt: off
            self.all_tag_values = list(
                self.main_window.pipeline_manager.pipelineEditorTabs.
                get_current_editor().all_tag_values_list
            )
            self.combo_box.addItems(
                self.main_window.pipeline_manager.pipelineEditorTabs.
                get_current_editor().tag_values_list
            )
            # fmt: on
            self.update_table()

    def update_table(self):
        """
        Update the iteration table.

        """

        # Updating the scan list
        if not self.scan_list:
            self.scan_list = self.project.session.get_documents_names(
                COLLECTION_CURRENT
            )

        # Clearing the table and preparing its columns
        self.iteration_table.clear()
        self.iteration_table.setColumnCount(len(self.push_buttons))

        # fmt: off
        if (
            self.main_window.pipeline_manager.pipelineEditorTabs.
            get_current_editor().iterated_tag
            is not None
        ):
            # fmt: on
            # Headers
            for idx in range(len(self.push_buttons)):
                # FIXME should not use GUI text values !!
                header_name = self.push_buttons[idx].text().replace("&", "")
                if header_name not in self.project.session.get_fields_names(
                    COLLECTION_CURRENT
                ):
                    print("{0} not in the project's tags".format(header_name))
                    return

                item = QTableWidgetItem()
                item.setText(header_name)
                self.iteration_table.setHorizontalHeaderItem(idx, item)

            # Searching the database scans that correspond to
            # iterated tag value
            # fmt: off
            filter_query = (
                "({"
                + str(
                    self.main_window.pipeline_manager.pipelineEditorTabs.
                    get_current_editor().iterated_tag
                )
                + "} "
                + "=="
                + ' "'
                + str(self.combo_box.currentText()).replace("&", "")
                + '")'
            )
            # fmt: on
            scans_list = self.project.session.filter_documents(
                COLLECTION_CURRENT, filter_query
            )
            scans_res = [
                getattr(document, TAG_FILENAME) for document in scans_list
            ]

            # Taking the intersection between the found database scans and the
            # user selection in the data_browser
            self.iteration_scans = list(
                set(scans_res).intersection(self.scan_list)
            )
            self.iteration_table.setRowCount(len(self.iteration_scans))

            # Filling the table cells
            row = -1
            for scan_name in self.iteration_scans:
                row += 1
                for idx in range(len(self.push_buttons)):
                    tag_name = self.push_buttons[idx].text().replace("&", "")

                    item = QTableWidgetItem()
                    item.setText(
                        str(
                            self.project.session.get_value(
                                COLLECTION_CURRENT, scan_name, tag_name
                            )
                        )
                    )
                    self.iteration_table.setItem(row, idx, item)

            all_iterations_scans = []

            # fmt: off
            for (
                tag_value
            ) in (
                self.main_window.pipeline_manager.pipelineEditorTabs.
                get_current_editor().tag_values_list
            ):
                # Searching the database scans that correspond to
                # iterated tag value
                filter_query = (
                    "({"
                    + str(
                        self.main_window.pipeline_manager.pipelineEditorTabs.
                        get_current_editor().iterated_tag
                    )
                    + "} "
                    + "=="
                    + ' "'
                    + str(tag_value)
                    + '")'
                )
                # fmt: on
                scans_list = self.project.session.filter_documents(
                    COLLECTION_CURRENT, filter_query
                )
                scans_res = [
                    getattr(document, TAG_FILENAME) for document in scans_list
                ]
                all_iterations_scans.append(
                    list(set(scans_res).intersection(self.scan_list))
                )
            self.all_iterations_scans = all_iterations_scans
            # self.scans = True

            # This will change the scans list in the current
            # Pipeline Manager tab
            self.iteration_table_updated.emit(
                self.iteration_scans, self.all_iterations_scans
            )

    def update_selected_tag(self, selected_tag):
        """Update the list of values corresponding to the selected tag

        :param selected_tag: the selected tag
        """

        tag_values_list = []
        scans_names = self.project.session.get_documents_names(
            COLLECTION_CURRENT
        )

        if not self.scan_list:
            self.scan_list = scans_names

        scans_names = list(set(scans_names).intersection(self.scan_list))

        for scan_name in scans_names:
            tag_value = self.project.session.get_value(
                COLLECTION_CURRENT, scan_name, selected_tag
            )

            if str(tag_value) not in tag_values_list:
                tag_values_list.append(str(tag_value))

        # fmt: off
        (
            self.main_window.pipeline_manager.pipelineEditorTabs.
            get_current_editor
        )().tag_values_list = tag_values_list

        (
            self.main_window.pipeline_manager.pipelineEditorTabs.
            get_current_editor
        )().all_tag_values_list = tag_values_list
        # fmt: on
        self.update_iterated_tag(selected_tag)



# -*- coding: utf-8 -*-
"""Module to define main window appearance, functions and settings.

Initialize the software appearance and defines interactions with the user.

:Contains:
    :Class:
        - MainWindow
        - _ProcDeleter

"""

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

import glob
import os
import shutil
import subprocess
import sys
import threading
import time
import webbrowser
from datetime import datetime
from os.path import expanduser

import yaml
from packaging import version
from PyQt5.QtCore import QCoreApplication, Qt

# PyQt5 imports
from PyQt5.QtGui import QCursor, QIcon
from PyQt5.QtWidgets import (
    QAction,
    QApplication,
    QMainWindow,
    QMenu,
    QMessageBox,
    QPushButton,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

import populse_mia.data_manager.data_loader as data_loader
from populse_mia.data_manager.project import (
    COLLECTION_CURRENT,
    TAG_HISTORY,
    Project,
)

# Populse_MIA imports
from populse_mia.data_manager.project_properties import SavedProjects
from populse_mia.software_properties import Config
from populse_mia.user_interface.data_browser.data_browser import DataBrowser
from populse_mia.user_interface.data_viewer.data_viewer_tab import (
    DataViewerTab,
)
from populse_mia.user_interface.pipeline_manager.pipeline_manager_tab import (
    PipelineManagerTab,
)
from populse_mia.user_interface.pipeline_manager.process_library import (
    InstallProcesses,
    PackageLibraryDialog,
)
from populse_mia.user_interface.pop_ups import (
    PopUpDeletedProject,
    PopUpDeleteProject,
    PopUpNewProject,
    PopUpOpenProject,
    PopUpPreferences,
    PopUpProperties,
    PopUpQuit,
    PopUpSaveProjectAs,
    PopUpSeeAllProjects,
)

CLINICAL_TAGS = [
    "Site",
    "Spectro",
    "MR",
    "PatientRef",
    "Pathology",
    "Age",
    "Sex",
    "Message",
]


console_shell_running = False
_ipsubprocs_lock = threading.RLock()
_ipsubprocs = []


class _ProcDeleter(threading.Thread):
    """Used by open_shell."""

    def __init__(self, o):
        threading.Thread.__init__(self)
        self.o = o

    def __del__(self):
        """Blabla"""

        try:
            self.o.kill()
        except Exception:
            pass
        if getattr(self, "console", False):
            global console_shell_running
            console_shell_running = False

    def run(self):
        """Blabla"""

        try:
            self.o.communicate()
        except Exception as e:
            print("exception in ipython process:", e)
        try:
            with _ipsubprocs_lock:
                _ipsubprocs.remove(self)
        except Exception:
            pass


class MainWindow(QMainWindow):
    """Initialize software appearance and define interactions with the user.

    .. Methods:
        - __init__ : initialise the object MainWindow
        - add_clinical_tags: add the clinical tags to the database and the
                             data browser
        - check_unsaved_modifications: check if there are differences
          between the current project and the database
        - check_database: check if files in database have been modified or
                          removed since they have been converted for the
                          first time
        - closeEvent: override the closing event to check if there are
          unsaved modifications
        - create_view_actions: create the actions in each menu
        - create_view_menus: create the menu-bar
        - create_view_window: create the main window view
        - create_project_pop_up: create a new project
        - create_tabs: create the tabs
        - credits: open the credits in a web browser
        - del_clinical_tags: Remove the clinical tags to the database and the
                             data browser
        - documentation: open the documentation in a web browser
        - get_controller_version: returns controller_version_changed attribute
        - import_data: call the import software (MRI File Manager)
        - install_processes_pop_up: open the install processes pop-up
        - open_project_pop_up: open a pop-up to open a project and updates
          the recent projects
        - open_recent_project: open a recent project
        - package_library_pop_up: open the package library pop-up
        - project_properties_pop_up: open the project properties pop-up
        - redo: redo the last action made by the user
        - remove_raw_files_useless: remove the useless raw files of the
          current project
        - save: save either the current project or the current pipeline
        - save_as: save either the current project or the current pipeline
          under a new name
        - save_project_as: open a pop-up to save the current project as
        - saveChoice: checks if the project needs to be saved as or just saved
        - see_all_projects: open a pop-up to show the recent projects
        - set_controller_version: Reverses controller_version_changed attribute
        - software_preferences_pop_up: open the Mia preferences pop-up
        - switch_project: switches project if it's possible
        - tab_changed: method called when the tab is changed
        - undo: undoes the last action made by the user
        - update_package_library_action: update the package library action
          depending on the mode
        - update_project: update the project once the database has been
          updated
        - update_recent_projects_actions: update the list of recent projects

    """

    def __init__(self, project, test=False, deleted_projects=None):
        """Main window class, initializes the software appearance and defines
        interactions with the user.

            :Parameter:
                - :project: current project in the software
                - :test: boolean if the widget is launched from unit tests
                     or not
                - :deleted_projects: projects that have been deleted

        """
        super(MainWindow, self).__init__()

        QApplication.restoreOverrideCursor()

        # We associate these methods and the instance to be able to call them
        # from anywhere.
        QCoreApplication.instance().title = self.windowTitle
        QCoreApplication.instance().set_title = self.setWindowTitle

        if deleted_projects is not None and deleted_projects:
            self.msg = PopUpDeletedProject(deleted_projects)

        self.config = Config()
        self.config.setSourceImageDir(
            os.path.join(
                os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
                "sources_images",
            )
        )
        self.windowName = "MIA - Multiparametric Image Analysis"
        self.projectName = "Unnamed project"
        self.project = project
        self.test = test

        self.saved_projects = SavedProjects()
        self.saved_projects_list = self.saved_projects.pathsList

        self.saved_projects_actions = []

        self.controller_version_changed = False

        # Define main window view
        self.create_view_window()

        # Initialize menu
        self.menu_file = self.menuBar().addMenu("File")
        self.menu_edition = self.menuBar().addMenu("Edit")
        self.menu_help = self.menuBar().addMenu("Help")
        self.menu_about = self.menuBar().addMenu("About")
        self.menu_more = self.menuBar().addMenu("More")
        self.menu_install_process = QMenu("Install processes", self)
        self.menu_saved_projects = QMenu("Saved projects", self)

        # Initialize tabs
        self.tabs = QTabWidget()
        self.data_browser = DataBrowser(self.project, self)
        self.data_viewer = DataViewerTab(self)
        self.pipeline_manager = PipelineManagerTab(self.project, [], self)
        self.centralWindow = QWidget()

        # Initialize menu actions
        sources_images_dir = Config().getSourceImageDir()
        self.action_save_project = self.menu_file.addAction("Save project")
        self.action_save_project_as = self.menu_file.addAction(
            "Save project as"
        )
        self.action_delete_project = self.menu_file.addAction("Delete project")
        self.action_create = QAction("New project", self)
        self.action_open = QAction("Open project", self)
        self.action_save = QAction("Save", self)
        self.action_save_as = QAction("Save as", self)
        self.action_delete = QAction("Delete project", self)
        self.action_import = QAction(
            QIcon(os.path.join(sources_images_dir, "Blue.png")), "Import", self
        )
        self.action_check_database = QAction("Check the whole database", self)
        self.action_see_all_projects = QAction("See all projects", self)
        self.action_project_properties = QAction("Project properties", self)
        self.action_software_preferences = QAction("MIA preferences", self)
        self.action_package_library = QAction("Package library manager", self)
        self.action_open_shell = QAction("Open python shell", self)
        self.action_exit = QAction(
            QIcon(os.path.join(sources_images_dir, "exit.png")), "Exit", self
        )
        self.action_undo = QAction("Undo", self)
        self.action_redo = QAction("Redo", self)
        self.action_documentation = QAction("Documentation", self)
        self.action_credits = QAction("Credits", self)
        self.action_install_processes_folder = QAction("From folder", self)
        self.action_install_processes_zip = QAction("From zip file", self)

        # Connect actions & menus views
        self.create_view_actions()
        self.create_view_menus()

        # Create Tabs
        self.create_tabs()
        self.setCentralWidget(self.centralWindow)
        if self.config.get_mainwindow_maximized():
            self.showMaximized()
        else:
            size = self.config.get_mainwindow_size()
            if size:
                self.resize(size[0], size[1])

    @staticmethod
    def last_window_closed():
        """Force exit the event loop after ipython console is closed.

        If the ipython console has been run, something prevents Qt from
        quitting after the window is closed. The cause is not known yet.
        So: force exit the event loop.
        """

        from soma.qt_gui.qt_backend import Qt

        Qt.QTimer.singleShot(10, Qt.qApp.exit)

    def add_clinical_tags(self):
        """Add the clinical tags to the database and the data browser"""

        added_tags = self.project.add_clinical_tags()

        for tag in added_tags:
            column = self.data_browser.table_data.get_index_insertion(tag)
            self.data_browser.table_data.add_column(column, tag)
            # self.project.unsavedModifications = True

    def check_unsaved_modifications(self):
        """Check if there are differences between the current project and the
        database.

        :return: Boolean. True if there are unsaved modifications,
           False otherwise
        """
        if self.project.isTempProject:
            if (
                len(
                    self.project.session.get_documents_names(
                        COLLECTION_CURRENT
                    )
                )
                > 0
            ):
                return True
            else:
                return False
        elif self.project.hasUnsavedModifications():
            return True
        else:
            return False

    def check_database(self):
        """Check if files in database have been modified since first import."""

        if self.project is None:
            return

        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))

        print("verify scans...")
        t0 = time.time()
        problem_list = data_loader.verify_scans(self.project)
        print("check time:", time.time() - t0)

        QApplication.restoreOverrideCursor()

        # Message if invalid files
        if problem_list:
            str_msg = ""
            for element in problem_list:
                str_msg += element + "\n\n"
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Warning)
            msg.setText(
                "These files have been modified or removed since "
                "they have been converted for the first time:"
            )
            msg.setInformativeText(str_msg)
            msg.setWindowTitle("Warning")
            msg.setStandardButtons(QMessageBox.Ok)
            msg.buttonClicked.connect(msg.close)
            msg.exec()

    def closeEvent(self, event):
        """Override the QWidget closing event to check if there are unsaved
        modifications

        :param event: closing event
        """
        if self.check_unsaved_modifications() is False or self.test:
            can_exit = True

        else:
            self.pop_up_close = PopUpQuit(self.project)
            self.pop_up_close.save_as_signal.connect(self.saveChoice)
            self.pop_up_close.exec()
            can_exit = self.pop_up_close.can_exit()

        if can_exit:
            if self.pipeline_manager.init_clicked:
                self.project.unsaveModifications()
                for brick in self.pipeline_manager.brick_list:
                    self.data_browser.table_data.delete_from_brick(brick)

            # Clean up
            config = Config()
            opened_projects = config.get_opened_projects()
            if self.project.folder in opened_projects:
                opened_projects.remove(self.project.folder)
            config.set_opened_projects(opened_projects)

            # Change controller version if needed
            if self.controller_version_changed:
                self.msg = QMessageBox()
                self.msg.setIcon(QMessageBox.Warning)
                self.msg.setText("Controller version change")
                self.msg.setInformativeText(
                    "A change of controller version, from {0} to {1}, "
                    "is planned for next start-up. Do you confirm that "
                    "you would like to perform this "
                    "change?".format(
                        "V1" if config.isControlV1() else "V2",
                        "V2" if config.isControlV1() else "V1",
                    )
                )
                self.msg.setWindowTitle("Warning")
                self.msg.setStandardButtons(
                    QMessageBox.Yes | QMessageBox.Cancel
                )
                return_value = self.msg.exec()

                if return_value == QMessageBox.Yes:
                    config.setControlV1(not config.isControlV1())
                self.msg.close()

            config.saveConfig()
            self.remove_raw_files_useless()

            event.accept()
        else:
            event.ignore()

        if self.data_browser.viewer:
            self.data_browser.viewer.clear()

        if self.data_viewer:
            self.data_viewer.clear()

    def create_view_actions(self):
        """Create the actions and their shortcuts in each menu"""

        self.action_create.setShortcut("Ctrl+N")
        self.action_open.setShortcut("Ctrl+O")
        self.action_save.setShortcut("Ctrl+S")
        self.addAction(self.action_save)

        self.action_save_as.setShortcut("Ctrl+Shift+S")
        self.addAction(self.action_save_as)
        self.addAction(self.action_delete)

        self.action_import.setShortcut("Ctrl+I")

        for i in range(self.config.get_max_projects()):
            self.saved_projects_actions.append(
                QAction(
                    self, visible=False, triggered=self.open_recent_project
                )
            )
        # if Config().get_user_mode() == True:
        #     self.action_package_library.setDisabled(True)
        # else:
        #     self.action_package_library.setEnabled(True)

        if Config().get_user_mode() is True:
            self.action_delete_project.setDisabled(True)
        else:
            self.action_delete_project.setEnabled(True)

        self.action_exit.setShortcut("Ctrl+W")

        self.action_undo.setShortcut("Ctrl+Z")

        self.action_redo.setShortcut("Ctrl+Y")

        # if Config().get_user_mode() == True:
        #     self.action_install_processes.setDisabled(True)
        # else:
        #     self.action_install_processes.setEnabled(True)

        # Connection of the several triggered signals of the actions to some
        # other methods
        self.action_create.triggered.connect(self.create_project_pop_up)
        self.action_open.triggered.connect(self.open_project_pop_up)
        self.action_exit.triggered.connect(self.close)
        self.action_check_database.triggered.connect(self.check_database)
        self.action_open_shell.triggered.connect(self.open_shell)
        self.action_save.triggered.connect(self.save)
        self.action_save_as.triggered.connect(self.save_as)
        self.action_delete.triggered.connect(self.delete_project)
        self.action_import.triggered.connect(self.import_data)
        self.action_see_all_projects.triggered.connect(self.see_all_projects)
        self.action_project_properties.triggered.connect(
            self.project_properties_pop_up
        )
        self.action_software_preferences.triggered.connect(
            self.software_preferences_pop_up
        )
        self.action_package_library.triggered.connect(
            self.package_library_pop_up
        )
        self.action_undo.triggered.connect(self.undo)
        self.action_redo.triggered.connect(self.redo)
        self.action_documentation.triggered.connect(self.documentation)
        self.action_credits.triggered.connect(self.credits)
        self.action_install_processes_folder.triggered.connect(
            lambda: self.install_processes_pop_up(folder=True)
        )
        self.action_install_processes_zip.triggered.connect(
            lambda: self.install_processes_pop_up(folder=False)
        )

    def create_view_menus(self):
        """Create the menu-bar view."""

        self.menu_more.addMenu(self.menu_install_process)

        # Actions in the "File" menu
        self.menu_file.addAction(self.action_create)
        self.menu_file.addAction(self.action_open)
        self.menu_file.addAction(self.action_check_database)

        self.action_save_project.triggered.connect(self.saveChoice)
        self.action_save_project_as.triggered.connect(self.save_project_as)
        self.action_delete_project.triggered.connect(self.delete_project)
        self.menu_file.addSeparator()
        self.menu_file.addAction(self.action_import)
        self.menu_file.addSeparator()
        self.menu_file.addMenu(self.menu_saved_projects)
        for i in range(self.config.get_max_projects()):
            self.menu_saved_projects.addAction(self.saved_projects_actions[i])
        self.menu_saved_projects.addSeparator()
        self.menu_saved_projects.addAction(self.action_see_all_projects)
        self.menu_file.addSeparator()
        self.menu_file.addAction(self.action_software_preferences)
        self.menu_file.addAction(self.action_project_properties)
        self.menu_file.addAction(self.action_package_library)
        self.menu_file.addSeparator()
        self.menu_file.addAction(self.action_open_shell)
        self.menu_file.addSeparator()
        self.menu_file.addAction(self.action_exit)
        self.update_recent_projects_actions()

        # Actions in the "Edition" menu
        self.menu_edition.addAction(self.action_undo)
        self.menu_edition.addAction(self.action_redo)

        # Actions in the "Help" menu
        self.menu_help.addAction(self.action_documentation)
        self.menu_help.addAction(self.action_credits)

        # Actions in the "More > Install processes" menu
        self.menu_install_process.addAction(
            self.action_install_processes_folder
        )
        self.menu_install_process.addAction(self.action_install_processes_zip)

    def create_view_window(self):
        """Create the main window view."""
        sources_images_dir = Config().getSourceImageDir()
        app_icon = QIcon(
            os.path.join(sources_images_dir, "Logo_populse_mia_LR.jpeg")
        )
        self.setWindowIcon(app_icon)
        background_color = self.config.getBackgroundColor()
        text_color = self.config.getTextColor()

        if not self.config.get_user_mode():
            self.windowName += " (Admin mode)"
        self.windowName += " - "

        self.setStyleSheet(
            "background-color:"
            + background_color
            + ";color:"
            + text_color
            + ";"
        )
        self.statusBar().showMessage(
            "Please create a new project (Ctrl+N) or "
            "open an existing project (Ctrl+O)"
        )

        self.setWindowTitle(self.windowName + self.projectName)

    def create_project_pop_up(self):
        """Create a new project."""

        if self.check_unsaved_modifications():
            self.pop_up_close = PopUpQuit(self.project)
            self.pop_up_close.save_as_signal.connect(self.saveChoice)
            self.pop_up_close.exec()
            can_switch = self.pop_up_close.can_exit()

        else:
            can_switch = True

        if can_switch:
            # Opens a pop-up when the 'New project' action is clicked and
            # updates the recent projects

            try:
                self.exPopup = PopUpNewProject()

            except Exception as e:
                print("\ncreate_project_pop_up: ", e)
                self.msg = QMessageBox()
                self.msg.setIcon(QMessageBox.Critical)
                self.msg.setText("Invalid projects folder path")
                self.msg.setInformativeText(
                    "The projects folder path in MIA preferences is invalid!"
                )
                self.msg.setWindowTitle("Error")
                yes_button = self.msg.addButton(
                    "Open MIA preferences", QMessageBox.YesRole
                )
                self.msg.addButton(QMessageBox.Ok)
                self.msg.exec()

                if self.msg.clickedButton() == yes_button:
                    self.software_preferences_pop_up()
                    self.msg.close()

                else:
                    self.msg.close()

            else:
                if self.exPopup.exec():
                    self.exPopup.get_filename(self.exPopup.selectedFiles())
                    file_name = self.exPopup.relative_path

                    # Removing the old project from the list of
                    # currently opened projects
                    config = Config()
                    opened_projects = config.get_opened_projects()
                    opened_projects.remove(self.project.folder)
                    config.set_opened_projects(opened_projects)
                    config.saveConfig()

                    self.remove_raw_files_useless()  # We remove the useless
                    # files from the old project

                    self.project = Project(self.exPopup.relative_path, True)

                    self.update_project(
                        file_name
                    )  # project updated everywhere

    def create_tabs(self):
        """Create the tabs and initializes the DataBrowser and PipelineManager
        classes."""
        self.config = Config()

        self.tabs.setAutoFillBackground(False)
        self.tabs.setStyleSheet("QTabBar{font-size:16pt;text-align: center}")
        self.tabs.setMovable(True)

        self.tabs.addTab(self.data_browser, "Data Browser")

        self.tabs.addTab(self.data_viewer, "Data Viewer")
        self.tabs.addTab(self.pipeline_manager, "Pipeline Manager")

        self.tabs.currentChanged.connect(self.tab_changed)
        vertical_layout = QVBoxLayout()
        vertical_layout.addWidget(self.tabs)

        self.centralWindow.setLayout(vertical_layout)

    def credits(self):
        """Open the credits in a web browser"""
        webbrowser.open(
            "https://github.com/populse/populse_mia/graphs/contributors"
        )

    def del_clinical_tags(self):
        """Remove the clinical tags to the database and the data browser"""

        removed_tags = self.project.del_clinical_tags()

        for tag in removed_tags:
            self.data_browser.table_data.removeColumn(
                self.data_browser.table_data.get_tag_column(tag)
            )
            # self.project.unsavedModifications = True

    def delete_project(self):
        """Open a pop-up to open a project and updates the recent projects."""

        try:
            self.exPopup = PopUpDeleteProject(self)

        except Exception as e:
            print("\ndelete_project: ", e)
            self.msg = QMessageBox()
            self.msg.setIcon(QMessageBox.Critical)
            self.msg.setText("Invalid projects folder path")
            self.msg.setInformativeText(
                "The projects folder path in MIA preferences is invalid!"
            )
            self.msg.setWindowTitle("Error")
            yes_button = self.msg.addButton(
                "Open MIA preferences", QMessageBox.YesRole
            )
            self.msg.addButton(QMessageBox.Ok)
            self.msg.exec()

            if self.msg.clickedButton() == yes_button:
                self.software_preferences_pop_up()
                self.msg.close()

            else:
                self.msg.close()

        else:
            self.exPopup.exec()

    @staticmethod
    def documentation():
        """Open the documentation in a web browser."""
        webbrowser.open(
            "https://populse.github.io/populse_mia/html/index.html"
        )

    def get_controller_version(self):
        """Gives the value of the controller_version_changed attribute.

        :return: Boolean
        """
        return self.controller_version_changed

    def install_processes_pop_up(self, folder=False):
        """Open the install processes pop-up.

        :param folder: boolean, True if installing from a folder

        """
        self.pop_up_install_processes = InstallProcesses(self, folder=folder)
        self.pop_up_install_processes.show()
        self.pop_up_install_processes.process_installed.connect(
            self.pipeline_manager.processLibrary.update_process_library
        )
        self.pop_up_install_processes.process_installed.connect(
            self.pipeline_manager.processLibrary.pkg_library.update_config
        )

    def import_data(self):
        """Call the import software (MRI File Manager), reads the imported
        files and loads them into the database.

        """
        # Opens the conversion software to convert the MRI files in Nifti/Json
        config = Config()
        home = expanduser("~")
        print("\nmri_conv opening ...\n")

        try:
            # Xmxsize: Specifies the maximum size (in bytes) of the memory
            #          allocation pool in bytes
            # Start with 4096M
            code_exit = subprocess.call(
                [
                    "java",
                    "-Xmx4096M",
                    "-jar",
                    config.get_mri_conv_path(),
                    "[ProjectsDir] " + home,
                    "[ExportNifti] "
                    + os.path.join(self.project.folder, "data", "raw_data"),
                    "[ExportToMIA] PatientName-StudyName-"
                    "CreationDate-SeqNumber-Protocol-"
                    "SequenceName-AcquisitionTime",
                    "CloseAfterExport",
                    "[ExportOptions] 00013",
                ]
            )

            if code_exit != 0 and code_exit != 100:
                raise ValueError("mri_conv did not run properly!")

        except ValueError:
            print(
                "\nMri_conv: Test with lower maximum heap "
                "size (4096M -> 1024M)...\n"
            )
            code_exit = subprocess.call(
                [
                    "java",
                    "-Xmx1024M",
                    "-jar",
                    config.get_mri_conv_path(),
                    "[ProjectsDir] " + home,
                    "[ExportNifti] "
                    + os.path.join(self.project.folder, "data", "raw_data"),
                    "[ExportToMIA] PatientName-StudyName-"
                    "CreationDate-SeqNumber-Protocol-"
                    "SequenceName-AcquisitionTime",
                    "CloseAfterExport",
                    "[ExportOptions] 00013",
                ]
            )

        # 'NoLogExport' if we don't want log export

        if code_exit == 0:
            # Database filled
            new_scans = data_loader.read_log(self.project, self)

            # Table updated
            documents = self.project.session.get_documents_names(
                COLLECTION_CURRENT
            )
            self.data_browser.table_data.scans_to_visualize = documents
            self.data_browser.table_data.scans_to_search = documents
            self.data_browser.table_data.add_columns()
            self.data_browser.table_data.fill_headers()
            self.data_browser.table_data.add_rows(new_scans)
            self.data_browser.reset_search_bar()
            self.data_browser.frame_advanced_search.setHidden(True)
            self.data_browser.advanced_search.rows = []
            self.project.unsavedModifications = True

        elif code_exit == 100:  # User only close mri_conv and do nothing
            pass

        else:
            print(
                "\nmri_conv, did not work properly. Current absolute"
                "path to MRIManager.jar defined in File > MIA Preferences:"
                "\n{0}\n".format(config.get_mri_conv_path())
            )

            if not os.path.isfile(config.get_mri_conv_path()):
                mssgText = (
                    "Warning: mri_conv did not work properly. The "
                    "current absolute path to MRIManager.jar doesn't "
                    "seem to be correctly defined.\nCurrent absolute "
                    "path to MRIManager.jar defined in\nFile > MIA "
                    "Preferences:\n{0}".format(config.get_mri_conv_path())
                )

            else:
                mssgText = (
                    "Warning : mri_conv did not work properly. Please "
                    "check if the currently installed mri_conv Java "
                    "ARchive is not corrupted.\nCurrent absolute path "
                    "to MRIManager.jar defined in\nFile > MIA "
                    "Preferences:\n{0}".format(config.get_mri_conv_path())
                )

            msg = QMessageBox()
            msg.setIcon(QMessageBox.Warning)
            msg.setWindowTitle("populse_mia - Warning: Data import issue!")
            msg.setText(mssgText)
            msg.setStandardButtons(QMessageBox.Ok)
            msg.buttonClicked.connect(msg.close)
            msg.exec()

    def open_project_pop_up(self):
        """Open a pop-up to open a project and updates the recent projects."""
        # Ui_Dialog() is defined in pop_ups.py
        # We check for unsaved modifications
        if self.check_unsaved_modifications():
            # If there are unsaved modifications, we ask the user what he
            # wants to do
            self.pop_up_close = PopUpQuit(self.project)
            self.pop_up_close.save_as_signal.connect(self.saveChoice)
            self.pop_up_close.exec()
            can_switch = self.pop_up_close.can_exit()

        else:
            can_switch = True

        # We can open a new project
        if can_switch:
            try:
                self.exPopup = PopUpOpenProject()

            except Exception as e:
                print("\nopen_project_pop_up: ", e)
                self.msg = QMessageBox()
                self.msg.setIcon(QMessageBox.Critical)
                self.msg.setText("Invalid projects folder path")
                self.msg.setInformativeText(
                    "The projects folder path in MIA preferences is invalid!"
                )
                self.msg.setWindowTitle("Error")
                yes_button = self.msg.addButton(
                    "Open MIA preferences", QMessageBox.YesRole
                )
                self.msg.addButton(QMessageBox.Ok)
                self.msg.exec()

                if self.msg.clickedButton() == yes_button:
                    self.software_preferences_pop_up()
                    self.msg.close()

                else:
                    self.msg.close()

            else:
                if self.exPopup.exec():
                    project_name = self.exPopup.selectedFiles()
                    self.exPopup.get_filename(project_name)
                    project_name = self.exPopup.relative_path
                    self.data_browser.data_sent = False

                    # We switch the project
                    self.switch_project(project_name, self.exPopup.name)
                    field_names = self.project.session.get_fields_names(
                        COLLECTION_CURRENT
                    )

                    if all(ele in field_names for ele in CLINICAL_TAGS):
                        Config().set_clinical_mode(True)

                    else:
                        Config().set_clinical_mode(False)

                    # Update the history and brick tables in the newly opened
                    # project, if it comes from outside.
                    path_name = os.path.dirname(
                        os.path.abspath(os.path.normpath(project_name))
                    )
                    projectsPath = os.path.abspath(
                        self.config.getPathToProjectsFolder()
                    )

                    if path_name != projectsPath:
                        self.project.update_db_for_paths(path_name)

    def open_recent_project(self):
        """Open a recent project."""
        # We check for unsaved modifications
        if self.check_unsaved_modifications():
            # If there are unsaved modifications, we ask the user what he
            # wants to do
            self.pop_up_close = PopUpQuit(self.project)
            self.pop_up_close.save_as_signal.connect(self.saveChoice)
            self.pop_up_close.exec()
            can_switch = self.pop_up_close.can_exit()

        else:
            can_switch = True

        # We can open a new project
        if can_switch:
            action = self.sender()
            if action:
                project_name = action.data()
                entire_path = os.path.abspath(project_name)
                path, name = os.path.split(entire_path)
                relative_path = os.path.relpath(project_name)
                self.switch_project(relative_path, name)
                # We switch the project
                field_names = self.project.session.get_fields_names(
                    COLLECTION_CURRENT
                )
                documents = self.project.session.get_documents_names(
                    COLLECTION_CURRENT
                )
                self.data_viewer.set_documents(self.project, documents)

                if all(ele in field_names for ele in CLINICAL_TAGS):
                    Config().set_clinical_mode(True)

                else:
                    Config().set_clinical_mode(False)

    def package_library_pop_up(self):
        """Open the package library pop-up"""

        self.pop_up_package_library = PackageLibraryDialog(
            mia_main_window=self
        )
        self.pop_up_package_library.setGeometry(300, 200, 800, 600)
        self.pop_up_package_library.show()
        self.pop_up_package_library.signal_save.connect(
            self.pipeline_manager.processLibrary.update_process_library
        )

    def project_properties_pop_up(self):
        """Open the project properties pop-up"""

        old_tags = self.project.session.get_shown_tags()
        self.pop_up_settings = PopUpProperties(
            self.project, self.data_browser, old_tags
        )
        self.pop_up_settings.setGeometry(300, 200, 800, 600)
        self.pop_up_settings.show()

        if self.pop_up_settings.exec():
            self.data_browser.table_data.update_visualized_columns(
                old_tags, self.project.session.get_shown_tags()
            )

    def redo(self):
        """Redo the last action made by the user."""
        if (
            self.tabs.tabText(self.tabs.currentIndex()).replace("&", "", 1)
            == "Data Browser"
        ):
            # In Data Browser
            self.project.redo(self.data_browser.table_data)
            # Action remade in the Database
        elif (
            self.tabs.tabText(self.tabs.currentIndex()).replace("&", "", 1)
            == "Pipeline Manager"
        ):
            # In Pipeline Manager
            self.pipeline_manager.redo()

    def remove_raw_files_useless(self):
        """Remove the useless raw files of the current project, close the
        database connection. The project is not valid any longer after this
        call."""

        folder = self.project.folder

        # If it's unnamed project, we can remove the whole project
        if self.project.isTempProject:
            # close database, and files
            self.project.session = None
            self.project.database.__exit__(None, None, None)
            self.project.database = None
            shutil.rmtree(folder)

        else:
            # I don't understand why files from raw_data are automatically
            # transferred to derived_data, if they are not in db. I comment
            # on this feature in the following lines. We can uncomment if
            # this action makes sense ...
            # for filename in glob.glob(
            #    os.path.join(os.path.abspath(folder), "data", "raw_data", "*")
            # ):
            #    scan = os.path.basename(filename)
            #    # The file is removed only if it's not a scan in the project,
            #    # and if it's not a logExport
            #    # Json files associated to nii files are kept for the raw_
            #    # data folder
            #    file_name, file_extension = os.path.splitext(scan)
            #    file_in_database = False
            #
            #    for database_scan in self.project.session.get_documents_names(
            #        COLLECTION_CURRENT
            #    ):
            #        if file_name in database_scan:
            #            file_in_database = True
            #
            #    if "logExport" in scan:
            #        file_in_database = True
            #
            #    if not file_in_database:
            #        os.rename(filename, filename.replace("raw_data",
            #                                             "derived_data"))

            # I don't understand why files from derived_data are automatically
            # deleted if they are not in db. I comment on this feature in the
            # following lines. We can uncomment if this action makes sense ...
            # for filename in glob.glob(
            #         os.path.join(os.path.relpath(
            #             self.project.folder), 'data', 'derived_data', '*')):
            #     scan = os.path.basename(filename)
            #     # The file is removed only if it's not a scan in the project,
            #     # and if it's not a logExport
            #     if (self.project.session.get_document(
            #             COLLECTION_CURRENT, os.path.join(
            #                 "data", "derived_data", scan)) is None and
            #             "logExport" not in scan):
            #         os.remove(filename)

            # I don't understand why files in downloaded_data are
            # automatically deleted if they are not in db. I comment on this
            # feature in the following line. We can uncomment if this action
            # makes sense...
            # for filename in glob.glob(
            #     os.path.join(
            #        os.path.abspath(self.project.folder),
            #        "data",
            #        "downloaded_data",
            #        "*",
            #    )
            # ):
            #    scan = os.path.basename(filename)
            #
            #    # The file is removed only if it's not a scan in the project,
            #    # and if it's not a logExport
            #    if (
            #        self.project.session.get_document(
            #            COLLECTION_CURRENT,
            #            os.path.join("data", "downloaded_data", scan),
            #        )
            #        is None
            #        and "logExport" not in scan
            #    ):
            #        os.remove(filename)
            #        self.project.unsavedModifications = True

            # close database, and files
            self.project.session = None
            self.project.database.__exit__(None, None, None)
            self.project.database = None

        self.project = None

    def save(self):
        """Save either the current project or the current pipeline"""

        if (
            self.tabs.tabText(self.tabs.currentIndex()).replace("&", "", 1)
            == "Data Browser"
        ):
            # In Data Browser
            self.saveChoice()
        elif (
            self.tabs.tabText(self.tabs.currentIndex()).replace("&", "", 1)
            == "Pipeline Manager"
        ):
            # In Pipeline Manager
            self.pipeline_manager.savePipeline()

    def save_as(self):
        """Save either the current project or the current pipeline under a new
        name.
        """
        if (
            self.tabs.tabText(self.tabs.currentIndex()).replace("&", "", 1)
            == "Data Browser"
        ):
            # In Data Browser
            self.save_project_as()
        elif (
            self.tabs.tabText(self.tabs.currentIndex()).replace("&", "", 1)
            == "Pipeline Manager"
        ):
            # In Pipeline Manager
            self.pipeline_manager.savePipelineAs()

    def save_project_as(self):
        """Open a pop-up to save the current project as"""
        try:
            self.exPopup = PopUpSaveProjectAs()

        except Exception as e:
            print("\nsave_project_as: ", e)
            self.msg = QMessageBox()
            self.msg.setIcon(QMessageBox.Critical)
            self.msg.setText("Invalid projects folder path")
            self.msg.setInformativeText(
                "The projects folder path in MIA preferences is invalid!"
            )
            self.msg.setWindowTitle("Error")
            yes_button = self.msg.addButton(
                "Open MIA preferences", QMessageBox.YesRole
            )
            self.msg.addButton(QMessageBox.Ok)
            self.msg.exec()

            if self.msg.clickedButton() == yes_button:
                self.software_preferences_pop_up()
                self.msg.close()

            else:
                self.msg.close()

        else:
            if self.test:
                self.exPopup.exec = lambda x=0: True
                self.exPopup.validate = True
                self.exPopup.new_project.text = lambda x=0: "something"
                self.exPopup.return_value()

            self.exPopup.exec()

            if self.exPopup.validate:
                old_folder_rel = self.project.folder
                old_folder = os.path.abspath(old_folder_rel)

                as_folder_rel = self.exPopup.relative_path
                as_folder = os.path.abspath(as_folder_rel)

                if as_folder_rel == old_folder_rel:
                    self.project.saveModifications()
                    return True

                database_path = os.path.join(as_folder, "database")
                properties_path = os.path.join(as_folder, "properties")
                filters_path = os.path.join(as_folder, "filters")
                data_path = os.path.join(as_folder, "data")

                raw_data_path = os.path.join(data_path, "raw_data")
                downloaded_data_path = os.path.join(
                    data_path, "downloaded_data"
                )

                # List of projects updated
                if not self.test:
                    self.saved_projects_list = (
                        self.saved_projects
                    ).addSavedProject(as_folder_rel)
                self.update_recent_projects_actions()

                if os.path.exists(as_folder_rel):
                    # Prevent by a careful message
                    # see PopUpSaveProjectAs/return_value
                    # in admin mode only
                    shutil.rmtree(as_folder_rel)

                if not os.path.exists(as_folder_rel):
                    os.makedirs(as_folder_rel)
                    os.mkdir(data_path)
                    os.mkdir(raw_data_path)
                    # os.mkdir(derived_data_path)
                    os.mkdir(downloaded_data_path)
                    os.mkdir(filters_path)

                # Data files copied
                if os.path.exists(os.path.join(old_folder_rel, "data")):
                    for filename in glob.glob(
                        os.path.join(old_folder, "data", "raw_data", "*")
                    ):
                        shutil.copy(
                            filename, os.path.join(data_path, "raw_data")
                        )

                    # for filename in glob.glob(
                    #     os.path.join(old_folder, "data", "derived_data", "*")
                    # ):
                    #     shutil.copy(
                    #         filename, os.path.join(data_path, "derived_data")
                    #     )
                    shutil.copytree(
                        os.path.join(old_folder, "data", "derived_data"),
                        os.path.join(data_path, "derived_data"),
                    )

                    for filename in glob.glob(
                        os.path.join(
                            old_folder, "data", "downloaded_data", "*"
                        )
                    ):
                        shutil.copy(
                            filename,
                            os.path.join(data_path, "downloaded_data"),
                        )

                if os.path.exists(os.path.join(old_folder_rel, "filters")):
                    for filename in glob.glob(
                        os.path.join(old_folder, "filters", "*")
                    ):
                        shutil.copy(filename, os.path.join(filters_path))

                # First we register the Database before committing the last
                # pending modifications
                shutil.copy(
                    os.path.join(old_folder, "database", "mia.db"),
                    os.path.join(
                        old_folder, "database", "mia_before_commit.db"
                    ),
                )

                # We commit the last pending modifications
                self.project.saveModifications()

                os.mkdir(properties_path)
                shutil.copy(
                    os.path.join(old_folder, "properties", "properties.yml"),
                    properties_path,
                )

                # We copy the Database with all the modifications committed in
                # the new project
                os.mkdir(database_path)
                shutil.copy(
                    os.path.join(old_folder, "database", "mia.db"),
                    database_path,
                )

                reset_old_db = not self.project.isTempProject

                # Removing the old project from the list of
                # currently opened projects
                config = Config()
                opened_projects = config.get_opened_projects()

                if self.project.folder in opened_projects:
                    opened_projects.remove(self.project.folder)
                config.set_opened_projects(opened_projects)
                config.saveConfig()

                # We remove the useless files from the old project
                self.remove_raw_files_useless()

                if reset_old_db:
                    # We remove the Database with all the modifications saved
                    # in the old project
                    os.remove(os.path.join(old_folder, "database", "mia.db"))

                    # We reput the Database without the last modifications
                    # in the old project
                    shutil.copy(
                        os.path.join(
                            old_folder, "database", "mia_before_commit.db"
                        ),
                        os.path.join(old_folder, "database", "mia.db"),
                    )

                    os.remove(
                        os.path.join(
                            old_folder, "database", "mia_before_commit.db"
                        )
                    )

                # project updated everywhere
                self.project = Project(as_folder_rel, False)
                self.project.setName(os.path.basename(as_folder_rel))
                self.project.setDate(
                    datetime.now().strftime("%d/%m/%Y %H:%M:%S")
                )
                self.project.saveModifications()

                self.update_project(as_folder_rel, call_update_table=False)
                # project updated everywhere

                # If some files have been set in the pipeline editors,
                # display a warning message
                if (
                    self.pipeline_manager.pipelineEditorTabs
                ).has_pipeline_nodes():
                    msg = QMessageBox()
                    msg.setIcon(QMessageBox.Warning)
                    msg.setText(
                        "This action moves the current database. "
                        "All pipelines will need to be initialized "
                        "again before they can run."
                    )
                    msg.setWindowTitle("Warning")
                    msg.setStandardButtons(QMessageBox.Ok)
                    msg.buttonClicked.connect(msg.close)
                    msg.exec()

        # Update of the history and the brick table in the newly
        # created project
        self.project.update_db_for_paths()

    def saveChoice(self):
        """Check if the project needs to be 'saved as' or just 'saved'."""
        if self.project.isTempProject:
            self.save_project_as()
        else:
            self.project.saveModifications()

    def see_all_projects(self):
        """Open a pop-up to show the recent projects."""
        # Ui_Dialog() is defined in pop_ups.py
        self.exPopup = PopUpSeeAllProjects(self.saved_projects, self)
        if self.exPopup.exec():
            file_path = self.exPopup.relative_path
            if not self.test:
                self.saved_projects_list = self.saved_projects.addSavedProject(
                    file_path
                )
            self.update_recent_projects_actions()

    def set_controller_version(self):
        """Reverses the value of the controller_version_changed attribute.

        From False to True and vice versa
        """
        self.controller_version_changed = not self.controller_version_changed

    def software_preferences_pop_up(self):
        """Open the Mia preferences pop-up."""

        self.pop_up_preferences = PopUpPreferences(self)
        self.pop_up_preferences.setGeometry(300, 200, 800, 600)
        self.pop_up_preferences.show()
        self.pop_up_preferences.use_clinical_mode_signal.connect(
            self.add_clinical_tags
        )
        self.pop_up_preferences.not_use_clinical_mode_signal.connect(
            self.del_clinical_tags
        )

        # Modifying the options in the Pipeline Manager
        # (verify if user mode)
        self.pop_up_preferences.signal_preferences_change.connect(
            self.pipeline_manager.update_user_mode
        )
        # self.pop_up_preferences.signal_preferences_change.connect(
        #     self.update_package_library_action)

    def open_shell(self):
        """Open a Qt console shell with an IPython kernel seeing the program
        internals
        """

        from soma.qt_gui import qt_backend

        ipfunc = None
        mode = "qtconsole"
        print("startShell")

        try:
            # to check it is installed
            import jupyter_core.application  # noqa: F401
            import qtconsole  # noqa: F401

            ipfunc = (
                "from jupyter_core import application; "
                "app = application.JupyterApp(); app.initialize(); app.start()"
            )

        except ImportError:
            print("failed to run Qt console")
            return

        if ipfunc:
            import soma.subprocess

            ipConsole = self.run_ipconsole_kernel(mode)

            if ipConsole:
                qt_api = qt_backend.get_qt_backend()
                qt_apis = {
                    "PyQt4": "pyqt",
                    "PyQt5": "pyqt5",
                    "PySide": "pyside",
                }
                qt_api_code = qt_apis.get(qt_api, "pyq5t")
                cmd = [
                    sys.executable,
                    "-c",
                    'import os; os.environ["QT_API"] = "%s"; %s'
                    % (qt_api_code, ipfunc),
                    mode,
                    "--existing",
                    "--shell=%d" % ipConsole.shell_port,
                    "--iopub=%d" % ipConsole.iopub_port,
                    "--stdin=%d" % ipConsole.stdin_port,
                    "--hb=%d" % ipConsole.hb_port,
                ]
                sp = soma.subprocess.Popen(cmd)
                pd = _ProcDeleter(sp)
                with _ipsubprocs_lock:
                    _ipsubprocs.append(pd)
                pd.start()

                # hack the lastWindowClosed event because it becomes inactive
                # otherwise
                QApplication.instance().lastWindowClosed.connect(
                    self.last_window_closed
                )

    @staticmethod
    def run_ipconsole_kernel(mode="qtconsole"):
        """blabla"""

        print("run_ipconsole_kernel:", mode)
        import IPython  # noqa: F401
        from IPython.lib import guisupport
        from soma.qt_gui.qt_backend import Qt

        qtapp = Qt.QApplication.instance()
        qtapp._in_event_loop = True
        guisupport.in_event_loop = True
        # ipversion = [int(x) for x in IPython.__version__.split(".")]

        from ipykernel.kernelapp import IPKernelApp

        app = IPKernelApp.instance()

        if not app.initialized() or not app.kernel:
            print("running IP console kernel")
            app.hb_port = 50042  # don't know why this is not set automatically
            app.initialize(
                [
                    mode,
                    "--gui=qt",  # '--pylab=qt',
                    "--KernelApp.parent_appname='ipython-%s'" % mode,
                ]
            )
            # in ipython >= 1.2, app.start() blocks until a ctrl-c is issued in
            # the terminal. Seems to block in tornado.ioloop.PollIOLoop.start()
            #
            # So, don't call app.start because it would begin a zmq/tornado
            # loop instead we must just initialize its callback.
            # if app.poller is not None:
            # app.poller.start()
            app.kernel.start()

            # IP 2 allows just calling the current callbacks.
            # For IP 1 it is not sufficient.
            import tornado
            from zmq.eventloop import ioloop

            if tornado.version_info >= (4, 5):
                # tornado 5 is using a decque for _callbacks, not a
                # list + explicit locking

                def my_start_ioloop_callbacks(self):
                    """Blabla"""

                    if hasattr(self, "_callbacks"):
                        ncallbacks = len(self._callbacks)
                        for i in range(ncallbacks):
                            self._run_callback(self._callbacks.popleft())

            else:

                def my_start_ioloop_callbacks(self):
                    """Blabla"""

                    with self._callback_lock:
                        callbacks = self._callbacks
                        self._callbacks = []

                    for callback in callbacks:
                        self._run_callback(callback)

            my_start_ioloop_callbacks(ioloop.IOLoop.instance())

        return app

    def switch_project(self, file_path, name):
        """Check if it's possible to open the selected project
        and quit the current one.

        :param file_path: raw file_path
        :param name: project name

        :return: Boolean
        """
        # /!\ file_path and path are the same param

        # Switching project only if it's a different one
        if file_path != self.project.folder:
            # If the file exists
            if os.path.exists(os.path.join(file_path)):
                # If it is a MIA project
                if (
                    os.path.exists(
                        os.path.join(file_path, "properties", "properties.yml")
                    )
                    and os.path.exists(
                        os.path.join(file_path, "database", "mia.db")
                    )
                    and os.path.exists(
                        os.path.join(file_path, "data", "raw_data")
                    )
                    and os.path.exists(
                        os.path.join(file_path, "data", "derived_data")
                    )
                    and os.path.exists(
                        os.path.join(file_path, "data", "downloaded_data")
                    )
                    and os.path.exists(os.path.join(file_path, "filters"))
                ):
                    # We check if the name of the project directory is the
                    # same in its properties
                    with open(
                        os.path.join(
                            file_path, "properties", "properties.yml"
                        ),
                        "r+",
                    ) as stream:
                        if version.parse(yaml.__version__) > version.parse(
                            "5.1"
                        ):
                            properties = yaml.load(
                                stream, Loader=yaml.FullLoader
                            )

                        else:
                            properties = yaml.load(stream)

                        path, name = os.path.split(file_path)

                        if properties["name"] != name:
                            properties["name"] = name
                            yaml.dump(
                                properties,
                                stream,
                                default_flow_style=False,
                                allow_unicode=True,
                            )

                    # We check for invalid scans in the project
                    try:
                        temp_database = Project(file_path, False)

                    except IOError:
                        msg = QMessageBox()
                        msg.setIcon(QMessageBox.Warning)
                        msg.setText("project already opened")
                        msg.setInformativeText(
                            "The project at "
                            + str(file_path)
                            + " is already opened in another "
                            "instance of the software."
                        )
                        msg.setWindowTitle("Warning")
                        msg.setStandardButtons(QMessageBox.Ok)
                        msg.buttonClicked.connect(msg.close)
                        msg.exec()
                        return False

                    # We check for valid version of the project

                    if not (temp_database.session.get_fields_names)(
                        COLLECTION_CURRENT
                    ) or (
                        TAG_HISTORY
                        not in (temp_database.session.get_fields_names)(
                            COLLECTION_CURRENT
                        )
                    ):
                        msg = QMessageBox()
                        msg.setIcon(QMessageBox.Warning)
                        msg.setText(
                            "The project cannot be read by Mia. Please check "
                            "if the version of the project is compatible with "
                            "the version of the running mia..."
                        )
                        msg.setWindowTitle("Warning")
                        msg.setStandardButtons(QMessageBox.Ok)
                        msg.buttonClicked.connect(msg.close)
                        msg.exec()
                        config = Config()
                        opened_projects = config.get_opened_projects()

                        if file_path in opened_projects:
                            opened_projects.remove(file_path)
                            config.set_opened_projects(opened_projects)
                            config.saveConfig()

                        return False

                    # project removed from the opened projects list
                    config = Config()
                    opened_projects = config.get_opened_projects()
                    if self.project.folder in opened_projects:
                        opened_projects.remove(self.project.folder)
                    config.set_opened_projects(opened_projects)
                    config.saveConfig()

                    # We remove the useless files from the old project
                    self.remove_raw_files_useless()

                    self.project = temp_database  # New Database

                    self.update_project(file_path)
                    # project updated everywhere

                    return True

                # Not a MIA project
                else:
                    msg = QMessageBox()
                    msg.setIcon(QMessageBox.Warning)
                    msg.setText(
                        "The project selected isn't a valid MIA project"
                    )
                    msg.setInformativeText(
                        "The project selected " + name + " isn't a MIA project"
                        ".\nPlease select a "
                        "valid one."
                    )
                    msg.setWindowTitle("Warning")
                    msg.setStandardButtons(QMessageBox.Ok)
                    msg.buttonClicked.connect(msg.close)
                    msg.exec()
                    return False

            # The project doesn't exist anymore
            else:
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Warning)
                msg.setText("The project selected doesn't exist anymore")
                msg.setInformativeText(
                    "The project selected " + name + " doesn't exist anymore."
                    "\nPlease select "
                    "another one."
                )
                msg.setWindowTitle("Warning")
                msg.setStandardButtons(QMessageBox.Ok)
                msg.buttonClicked.connect(msg.close)
                msg.exec()
                return False

    def tab_changed(self):
        """Update the window when the tab is changed."""

        if (
            self.tabs.tabText(self.tabs.currentIndex()).replace("&", "", 1)
            == "Data Browser"
        ):
            # data_browser refreshed after working with pipelines
            old_scans = self.data_browser.table_data.scans_to_visualize
            documents = self.project.session.get_documents_names(
                COLLECTION_CURRENT
            )

            self.data_browser.table_data.add_columns()
            self.data_browser.table_data.fill_headers()

            self.data_browser.table_data.add_rows(documents)

            self.data_browser.table_data.scans_to_visualize = documents
            self.data_browser.table_data.scans_to_search = documents

            self.data_browser.table_data.itemChanged.disconnect()
            self.data_browser.table_data.fill_cells_update_table()
            self.data_browser.table_data.itemChanged.connect(
                self.data_browser.table_data.change_cell_color
            )

            self.data_browser.table_data.update_visualized_rows(old_scans)

            # Advanced search + search_bar opened
            old_search = self.project.currentFilter.search_bar
            self.data_browser.reset_search_bar()
            self.data_browser.search_bar.setText(old_search)

            if len(self.project.currentFilter.nots) > 0:
                self.data_browser.frame_advanced_search.setHidden(False)
                self.data_browser.advanced_search.scans_list = (
                    self.data_browser.table_data.scans_to_visualize
                )
                self.data_browser.advanced_search.show_search()
                self.data_browser.advanced_search.apply_filter(
                    self.project.currentFilter
                )

        elif (
            self.tabs.tabText(self.tabs.currentIndex()).replace("&", "", 1)
            == "Data Viewer"
        ):
            self.data_viewer.load_viewer(self.data_viewer.current_viewer())
            documents = self.project.session.get_documents_names(
                COLLECTION_CURRENT
            )
            self.data_viewer.set_documents(self.project, documents)

        elif (
            self.tabs.tabText(self.tabs.currentIndex()).replace("&", "", 1)
            == "Pipeline Manager"
        ):
            if self.data_browser.data_sent is False:
                scans = self.project.session.get_documents_names(
                    COLLECTION_CURRENT
                )
                self.pipeline_manager.scan_list = scans
                self.pipeline_manager.nodeController.scan_list = scans
                self.pipeline_manager.pipelineEditorTabs.scan_list = scans
            self.pipeline_manager.pipelineEditorTabs.update_scans_list()
            self.pipeline_manager.update_user_buttons_states()

            # fmt: off
            if (
                self.pipeline_manager.pipelineEditorTabs.
                    get_current_editor().iterated_tag
            ):
                self.pipeline_manager.iterationTable.update_iterated_tag(
                    self.pipeline_manager.pipelineEditorTabs.
                    get_current_editor().iterated_tag
                )
            # fmt: on

            # Pipeline Manager
            # The pending modifications must be saved before
            # working with pipelines (auto_commit)
            if self.project.hasUnsavedModifications():
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Warning)
                msg.setText("Unsaved modifications in the Data Browser !")
                msg.setInformativeText(
                    "There are unsaved modifications in the database, "
                    "you need to save or remove them before working "
                    "with pipelines."
                )
                msg.setWindowTitle("Warning")
                save_button = QPushButton("Save")
                save_button.clicked.connect(self.project.saveModifications)
                unsave_button = QPushButton("Not Save")
                unsave_button.clicked.connect(self.project.unsaveModifications)
                msg.addButton(save_button, QMessageBox.AcceptRole)
                msg.addButton(unsave_button, QMessageBox.AcceptRole)
                msg.exec()

    def undo(self):
        """Undo the last action made by the user."""
        if (
            self.tabs.tabText(self.tabs.currentIndex()).replace("&", "", 1)
            == "Data Browser"
        ):
            # In Data Browser
            self.project.undo(self.data_browser.table_data)
            # Action reverted in the Database
        elif (
            self.tabs.tabText(self.tabs.currentIndex()).replace("&", "", 1)
            == "Pipeline Manager"
        ):
            # In Pipeline Manager
            self.pipeline_manager.undo()

    # def update_package_library_action(self):
    #     """Update the package library action depending on the mode."""
    #     if Config().get_user_mode() == True:
    #         self.action_package_library.setDisabled(True)
    #         # self.action_install_processes.setDisabled(True)
    #     else:
    #         self.action_package_library.setEnabled(True)
    #         # self.action_install_processes.setEnabled(True)

    def update_project(self, file_path, call_update_table=True):
        """Update the project once the database has been updated.
        Update the database, the window title and the recent and saved
        projects menus.

        :param file_path: File name of the new project
        :param call_update_table: boolean, True if we need to call
        """

        self.data_browser.update_database(self.project)

        # Database update data_browser
        self.pipeline_manager.update_project(self.project)

        if call_update_table:
            self.data_browser.table_data.update_table()  # Table updated

        # Window name updated
        if self.project.isTempProject:
            self.projectName = "Unnamed project"
        else:
            self.projectName = self.project.getName()
        self.setWindowTitle(self.windowName + self.projectName)

        # List of project updated
        if not self.test and not self.project.isTempProject:
            self.saved_projects_list = self.saved_projects.addSavedProject(
                file_path
            )
        self.update_recent_projects_actions()

    def update_recent_projects_actions(self):
        """Update the list of recent projects."""
        for j in range(0, self.config.get_max_projects()):
            self.saved_projects_actions[j].setVisible(False)
        if self.saved_projects_list:
            if len(self.saved_projects_list) > 0:
                for i in range(
                    min(
                        len(self.saved_projects_list),
                        self.config.get_max_projects(),
                    )
                ):
                    text = os.path.basename(self.saved_projects_list[i])
                    self.saved_projects_actions[i].setText(text)
                    self.saved_projects_actions[i].setData(
                        self.saved_projects_list[i]
                    )
                    self.saved_projects_actions[i].setVisible(True)



# -*- coding: utf-8 -*-
"""Module to handle the node of a pipeline and its plugs.

:Contains:
    :Class:
        - PlugFilter (must be declared before AttributesFilter)
        - AttributesFilter
        - CapsulNodeController
        - FilterWidget
        - NodeController

"""

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

import os
import sys
from functools import partial

import sip

# capsul imports
from capsul.attributes.completion_engine import ProcessCompletionEngine
from capsul.pipeline.pipeline_nodes import PipelineNode
from capsul.pipeline.process_iteration import ProcessIteration
from capsul.qt_gui.widgets.attributed_process_widget import (
    AttributedProcessWidget,
)
from matplotlib.backends.qt_compat import QtWidgets

# PyQt5 imports
from PyQt5 import Qt
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (
    QApplication,
    QDialog,
    QDialogButtonBox,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

# soma-base imports
from soma.controller import trait_ids
from traits.api import TraitError, Undefined

# Populse_MIA imports
from populse_mia.data_manager.filter import Filter
from populse_mia.data_manager.project import (
    COLLECTION_BRICK,
    COLLECTION_CURRENT,
    TAG_FILENAME,
)
from populse_mia.software_properties import Config
from populse_mia.user_interface.data_browser.advanced_search import (
    AdvancedSearch,
)
from populse_mia.user_interface.data_browser.data_browser import (
    TableDataBrowser,
    not_defined_value,
)
from populse_mia.user_interface.data_browser.rapid_search import RapidSearch
from populse_mia.user_interface.pipeline_manager.process_mia import ProcessMIA
from populse_mia.user_interface.pop_ups import (
    PopUpSelectTagCountTable,
    PopUpVisualizedTags,
)

from . import type_editors

if sys.version_info[0] >= 3:
    unicode = str

    def values(d):
        """blabla"""

        return list(d.values())

else:

    def values(d):
        """blabla"""

        return d.values()


class PlugFilter(QWidget):
    """Filter widget used on a node plug.

    The widget displays a browser with the selected files of the database,
    a rapid search and an advanced search to filter these files. Once the
    filtering is done, the result (as a list of files) is set to the plug.

    .. Methods:
        - ok_clicked: set the new value to the node plug and closes the widget
        - reset_search_bar: reset the search bar of the rapid search
        - search_str: update the files to display in the browser
        - set_plug_value: emit a signal to set the file names to the node plug
        - update_tag_to_filter: update the tag to Filter
        - update_tags: update the list of visualized tags
    """

    plug_value_changed = pyqtSignal(list)

    def __init__(
        self,
        project,
        scans_list,
        process,
        node_name,
        plug_name,
        node_controller,
        main_window,
    ):
        """
        Initialization of the PlugFilter widget

        :param project: current project in the software
        :param scans_list: list of database files to filter
        :param process: process instance of the selected node
        :param node_name: name of the current node
        :param plug_name: name of the selected node plug
        :param node_controller: parent node controller
        :param main_window: parent main window
        """
        super(PlugFilter, self).__init__(None)

        from populse_mia.data_manager.project import COLLECTION_CURRENT
        from populse_mia.user_interface.data_browser.rapid_search import (
            RapidSearch,
        )

        self.project = project
        self.node_controller = node_controller
        self.main_window = main_window
        self.process = process
        self.plug_name = plug_name

        # If the filter is saved in the node plug (not the case now)
        # if hasattr(self.process, 'filters'):
        #    if self.plug_name in self.process.filters.keys():
        #         print("Already a filter for {0} plug of {1} process".format(
        #                 self.plug_name, self.process.name))
        #         # TODO: fill the advanced search with the corresponding
        #          filter:orphan:

        # Verifying that the scan names begin not with a "/" or a "\"
        doc_list = []
        for brick in self.main_window.pipeline_manager.brick_list:
            doc = self.project.session.get_document(COLLECTION_BRICK, brick)
            if doc is not None:
                for key in doc["Output(s)"]:
                    if isinstance(doc["Output(s)"][key], str):
                        if doc["Output(s)"][key] != "":
                            doc_delete = os.path.relpath(
                                doc["Output(s)"][key], self.project.folder
                            )
                            doc_list.append(doc_delete)

        if scans_list:
            scans_list_copy = []
            for scan in scans_list:
                scan_no_pfolder = scan.replace(self.project.folder, "")
                if scan_no_pfolder[0] in ["\\", "/"]:
                    scan_no_pfolder = scan_no_pfolder[1:]
                if scan_no_pfolder not in doc_list:
                    scans_list_copy.append(scan_no_pfolder)

            self.scans_list = scans_list_copy

        # If there is no element in scans_list, this means that all the scans
        # of the database needs to be taken into account
        else:
            self.scans_list = self.project.session.get_documents_names(
                COLLECTION_CURRENT
            )

        self.setWindowTitle("Filter - " + node_name + " - " + plug_name)

        # Graphical components
        self.table_data = TableDataBrowser(
            self.project,
            self,
            self.node_controller.visibles_tags,
            False,
            True,
            link_viewer=False,
        )

        # Reducing the list of scans to selection
        all_scans = self.table_data.scans_to_visualize
        self.table_data.scans_to_visualize = self.scans_list
        self.table_data.scans_to_search = self.scans_list
        self.table_data.update_visualized_rows(all_scans)

        search_bar_layout = QHBoxLayout()

        self.rapid_search = RapidSearch(self)
        self.rapid_search.textChanged.connect(partial(self.search_str))

        sources_images_dir = Config().getSourceImageDir()
        self.button_cross = QToolButton()
        self.button_cross.setStyleSheet("background-color:rgb(255, 255, 255);")
        self.button_cross.setIcon(
            QIcon(os.path.join(sources_images_dir, "gray_cross.png"))
        )
        self.button_cross.clicked.connect(self.reset_search_bar)

        search_bar_layout.addWidget(self.rapid_search)
        search_bar_layout.addWidget(self.button_cross)

        self.advanced_search = AdvancedSearch(
            self.project,
            self,
            self.scans_list,
            self.node_controller.visibles_tags,
            from_pipeline=True,
        )
        self.advanced_search.show_search()

        push_button_tags = QPushButton("Visualized tags")
        push_button_tags.clicked.connect(self.update_tags)

        self.push_button_tag_filter = QPushButton(TAG_FILENAME)
        self.push_button_tag_filter.clicked.connect(self.update_tag_to_filter)

        push_button_ok = QPushButton("OK")
        push_button_ok.clicked.connect(self.ok_clicked)

        push_button_cancel = QPushButton("Cancel")
        push_button_cancel.clicked.connect(self.close)

        # Layout
        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(push_button_tags)
        buttons_layout.addWidget(self.push_button_tag_filter)
        buttons_layout.addStretch(1)
        buttons_layout.addWidget(push_button_ok)
        buttons_layout.addWidget(push_button_cancel)

        main_layout = QVBoxLayout()
        main_layout.addLayout(search_bar_layout)
        main_layout.addWidget(self.advanced_search)
        main_layout.addWidget(self.table_data)
        main_layout.addLayout(buttons_layout)

        self.setLayout(main_layout)

        screen_resolution = QApplication.instance().desktop().screenGeometry()
        width, height = screen_resolution.width(), screen_resolution.height()
        self.setMinimumWidth(round(0.6 * width))
        self.setMinimumHeight(round(0.8 * height))

    def ok_clicked(self):
        """Set the new value to the node plug and closes the widget."""

        # To use if the filters are set on plugs, which is not the case
        # if isinstance(self.process, ProcessMIA):
        #     (fields, conditions, values, links, nots) =
        #     self.advanced_search.get_filters(False)
        #
        #
        #     plug_filter = Filter(None, nots, values, fields,
        #     links, conditions, "")
        #     self.process.filters[self.plug_name] = plug_filter

        self.set_plug_value()
        self.close()

    def reset_search_bar(self):
        """Reset the search bar of the rapid search."""
        self.rapid_search.setText("")
        self.advanced_search.rows = []
        self.advanced_search.show_search()

        # All rows reput
        old_scan_list = self.table_data.scans_to_visualize
        self.table_data.scans_to_visualize = self.scans_list
        self.table_data.scans_to_search = self.scans_list
        self.table_data.update_visualized_rows(old_scan_list)

    def search_str(self, str_search):
        """Update the files to display in the browser.
        :param str_search: string typed in the rapid search
        """
        old_scan_list = self.table_data.scans_to_visualize

        # Every scan taken if empty search
        if str_search == "":
            return_list = self.table_data.scans_to_search
        else:
            # Scans with at least a not defined value
            if str_search == not_defined_value:
                filter = self.prepare_not_defined_filter(
                    self.project.session.get_shown_tags()
                )

            # Scans matching the search
            else:
                filter = self.rapid_search.prepare_filter(
                    str_search,
                    self.project.session.get_shown_tags(),
                    self.table_data.scans_to_search,
                )

            generator = self.project.session.filter_documents(
                COLLECTION_CURRENT, filter
            )

            # Creating the list of scans
            return_list = [getattr(scan, TAG_FILENAME) for scan in generator]

        self.table_data.scans_to_visualize = return_list
        self.advanced_search.scans_list = return_list

        # Rows updated
        self.table_data.update_visualized_rows(old_scan_list)

    def set_plug_value(self):
        """Emit a signal to set the file names to the node plug."""

        result_names = []
        points = self.table_data.selectedIndexes()

        # If the user has selected some items
        if points:
            for point in points:
                row = point.row()
                tag_name = self.push_button_tag_filter.text()
                if tag_name.startswith("&"):
                    tag_name = tag_name[1:]
                # We get the FileName of the scan from the first row
                scan_name = self.table_data.item(row, 0).text()
                value = self.project.session.get_value(
                    COLLECTION_CURRENT, scan_name, tag_name
                )
                if tag_name == TAG_FILENAME:
                    value = os.path.abspath(
                        os.path.join(self.project.folder, value)
                    )
                result_names.append(value)
        else:
            filter = self.table_data.get_current_filter()
            for i in range(len(filter)):
                scan_name = filter[i]
                tag_name = self.push_button_tag_filter.text()
                value = self.project.session.get_value(
                    COLLECTION_CURRENT, scan_name, tag_name
                )
                if tag_name == TAG_FILENAME:
                    value = os.path.abspath(
                        os.path.join(self.project.folder, value)
                    )
                result_names.append(value)

        self.plug_value_changed.emit(result_names)

    def update_tag_to_filter(self):
        """Update the tag to Filter."""

        popUp = PopUpSelectTagCountTable(
            self.project,
            self.node_controller.visibles_tags,
            self.push_button_tag_filter.text(),
        )
        if popUp.exec_():
            self.push_button_tag_filter.setText(popUp.selected_tag)

    def update_tags(self):
        """Update the list of visualized tags."""

        dialog = QDialog()
        visualized_tags = PopUpVisualizedTags(
            self.project, self.node_controller.visibles_tags
        )
        layout = QVBoxLayout()
        layout.addWidget(visualized_tags)
        buttons_layout = QHBoxLayout()
        buttons = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        buttons.accepted.connect(dialog.accept)
        buttons.rejected.connect(dialog.reject)
        buttons_layout.addWidget(buttons)
        layout.addLayout(buttons_layout)
        dialog.setLayout(layout)
        dialog.show()
        dialog.setMinimumHeight(600)
        dialog.setMinimumWidth(600)
        if dialog.exec():
            new_visibilities = []
            for x in range(visualized_tags.list_widget_selected_tags.count()):
                visible_tag = visualized_tags.list_widget_selected_tags.item(
                    x
                ).text()
                new_visibilities.append(visible_tag)
            new_visibilities.append(TAG_FILENAME)
            self.table_data.update_visualized_columns(
                self.node_controller.visibles_tags, new_visibilities
            )
            self.node_controller.visibles_tags = new_visibilities
            for row in self.advanced_search.rows:
                fields = row[2]
                fields.clear()
                for visible_tag in new_visibilities:
                    fields.addItem(visible_tag)
                fields.model().sort(0)
                fields.addItem("All visualized tags")


class AttributesFilter(PlugFilter):
    """Filter widget used on an attributes set for completion.

    The widget displays a browser with the selected files of the database,
    a rapid search and an advanced search to filter these files. Once the
    filtering is done, the result (as a list of files) is set to the plug.

    .. Methods:
        - ok_clicked: close the widget
    """

    attributes_selected = pyqtSignal(dict)

    def ok_clicked(self):
        """Close the widget"""
        self.close()
        attributes = {}

        points = self.table_data.selectedIndexes()

        # If the user has selected some items
        if points:
            for point in points:
                row = point.row()
                for tag_name in self.project.session.get_fields_names(
                    COLLECTION_CURRENT
                ):
                    # We get the FileName of the scan from the first row
                    scan_name = self.table_data.item(row, 0).text()
                    value = self.project.session.get_value(
                        COLLECTION_CURRENT, scan_name, tag_name
                    )
                    attributes.setdefault(tag_name, []).append(value)
        else:
            filter = self.table_data.get_current_filter()
            for i in range(len(filter)):
                scan_name = filter[i]
                for tag_name in self.project.session.get_fields_names(
                    COLLECTION_CURRENT
                ):
                    value = self.project.session.get_value(
                        COLLECTION_CURRENT, scan_name, tag_name
                    )
                    attributes.setdefault(tag_name, []).append(value)

        self.attributes_selected.emit(attributes)


# Node controller V2 style
class CapsulNodeController(QWidget):
    """
    Implementation of NodeController using Capsul AttributedProcessWidget
    widget.

    .. Methods:
        - display_parameters: display the parameters of the selected node
        - static_release: remove notification
        - release_process: remove notification from process
        - update_parameters: update the parameters values
        - parameters_changed: emit the value_changed signal
        - update_node_name: change the name of the selected node and updates
        the pipeline
        - rename_subprocesses: change the name of a node
        - filter_attributes: display a filter widget
        - update_attributes_from_filter: update attributes from filter widget
    """

    value_changed = pyqtSignal(list)

    def __init__(self, project, scan_list, pipeline_manager_tab, main_window):
        super().__init__()
        self.project = project
        self.scan_list = scan_list
        self.main_window = main_window
        self.node_name = ""
        self.visibles_tags = []
        self.pipeline = (
            pipeline_manager_tab.pipelineEditorTabs.get_current_pipeline()
        )

        # Layouts
        v_box_final = QVBoxLayout()
        self.setLayout(v_box_final)
        self.process_widget = None

        # Node name
        hlayout = QHBoxLayout()
        label_node_name = QLabel()
        label_node_name.setText("Node name:")

        self.line_edit_node_name = QLineEdit()
        hlayout.addWidget(label_node_name)
        hlayout.addWidget(self.line_edit_node_name)
        v_box_final.addLayout(hlayout)

    def display_parameters(self, node_name, process, pipeline):
        """Display the parameters of the selected node.

        The node parameters are read and line labels/line edits/push buttons
        are created for each of them. This methods consists mainly in widget
        and layout organization.

        :param node_name: name of the node
        :param process: process of the node
        :param pipeline: current pipeline
        """
        self.node_name = node_name
        self.pipeline = pipeline

        # The pipeline global inputs and outputs node name cannot be modified
        if self.node_name not in ("inputs", "outputs"):
            self.line_edit_node_name.setText(self.node_name)
            self.line_edit_node_name.setReadOnly(False)
            self.line_edit_node_name.returnPressed.connect(
                self.update_node_name
            )
        else:
            self.line_edit_node_name.setText("Pipeline inputs/outputs")
            self.line_edit_node_name.setReadOnly(True)

        if self.process_widget:
            # item = self.layout().takeAt(1)
            self.static_release(
                self.process_widget.attributed_process, self.parameters_changed
            )
            self.process_widget.deleteLater()
            # del item
            self.process_widget = None

        # get the list of inputs connected from outputs of upstream nodes
        # in order to disable their "filter" file button
        node = pipeline.nodes.get(node_name)
        connected_inputs = set()
        if node is not None:
            for plug_name, plug in node.plugs.items():
                if not plug.output and plug.links_from:
                    connected_inputs.add(plug_name)

        userlevel = Config().get_user_level()
        self.process = process

        # force initializing the completion engine
        ProcessCompletionEngine.get_completion_engine(process)
        # fmt: off
        self.process_widget = AttributedProcessWidget(
            process,
            override_control_types={
                "File": type_editors.PopulseFileControlWidget,
                "Directory": type_editors.PopulseDirectoryControlWidget,
                "List_File": type_editors.
                PopulseOffscreenListFileControlWidget,
                "Undefined": type_editors.PopulseUndefinedControlWidget,
            },
            separate_outputs=True,
            user_data={
                "project": self.project,
                "scan_list": self.scan_list,
                "main_window": self.main_window,
                "node_controller": self,
                "connected_inputs": connected_inputs,
            },
            scroll=False,
            userlevel=userlevel,
        )
        # fmt: on

        if hasattr(process, "completion_engine"):
            compl = process.completion_engine
            atts = compl.get_attribute_values()

            if len(atts.user_traits()) != 0:
                btn = QPushButton("Filter")
                btn.setSizePolicy(Qt.QSizePolicy.Fixed, Qt.QSizePolicy.Fixed)
                btn.clicked.connect(self.filter_attributes)
                self.process_widget.attrib_widget.layout().insertWidget(0, btn)

        self.layout().addWidget(self.process_widget)
        self.process.on_trait_change(self.parameters_changed, dispatch="ui")
        # this cannot be done in __del__ since the C++ part will be already
        # destroyed by then.
        # However this signal seems never to be emitted, and I don't understand
        # why. So release_process() has to be called manually from the
        # pipeline manager. Sigh.
        # self.process_widget.destroyed.connect(partial(
        #                                             self.static_release,
        #                                             process,
        #                                             self.parameters_changed))

    @staticmethod
    def static_release(process, param_changed):
        """Remove notification"""
        process.on_trait_change(param_changed, remove=True)

    def release_process(self):
        """Remove notification from process"""
        if hasattr(self, "process"):
            self.process.on_trait_change(self.parameters_changed, remove=True)
        try:
            if not sip.isdeleted(self):
                self.value_changed.disconnect()
        except TypeError:
            pass  # it was not connected: OK

    def update_parameters(self, process=None):
        """Update the parameters values.

        Does nothing any longer since the controller widget already reacts to
        changes in the process parameters.

        :param process: process of the node
        """
        pass

    def parameters_changed(self, _, plug_name, old_value, new_value):
        """Emit the value_changed signal."""

        # plug_name_type = type(plug_name)
        plug_type = type(new_value)
        self.value_changed.emit(
            [
                "plug_value",
                self.node_name,
                old_value,
                plug_name,
                plug_type,
                new_value,
            ]
        )

    def update_node_name(
        self,
        new_node_name=None,
        old_node_name=None,
        from_undo=False,
        from_redo=False,
    ):
        """Change the name of the selected node and updates the pipeline.

        Because the nodes are stored in a dictionary, we have to create
        a new node that has the same traits as the selected one and create
        new links that are the same than the selected node.

        :param new_node_name: new node name (is None except when this method
         is called from an undo/redo)
        :param old_node_name: old node name (is None except when this method
         is called from an undo/redo)
        :param from_undo: boolean, True if the action has been made using an
         undo
        :param from_redo: boolean, True if the action has been made using a
         redo
        """
        if not new_node_name:
            new_node_name = self.line_edit_node_name.text()

        if not old_node_name:
            old_node_name = self.node_name

        if isinstance(self.process, ProcessIteration):
            if not new_node_name.startswith("iterated_"):
                new_node_name = "iterated_" + new_node_name
                self.line_edit_node_name.setText(new_node_name)

        if new_node_name in list(self.pipeline.nodes.keys()):
            print(
                "\nUpdate of the node name from '{0}' to '{1}': impossible, "
                "the node '{1}' already exists !".format(
                    old_node_name, new_node_name
                )
            )

        else:
            self.pipeline.rename_node(old_node_name, new_node_name)
            self.rename_subprocesses(
                self.pipeline.nodes[new_node_name], new_node_name
            )

            # Updating the node_name attribute
            self.node_name = new_node_name
            self.pipeline.update_nodes_and_plugs_activation()

            # To undo/redo
            self.value_changed.emit(
                [
                    "node_name",
                    self.pipeline.nodes[new_node_name],
                    new_node_name,
                    old_node_name,
                ]
            )

            # For history
            history_maker = ["update_node_name"]

            if from_undo:
                # TODO: next line is strange!
                history_maker.append

            else:
                history_maker.append(self.pipeline.nodes[new_node_name])
                history_maker.append(new_node_name)
                history_maker.append(old_node_name)
            # fmt:off
            (
                self.main_window.pipeline_manager.pipelineEditorTabs.
                get_current_editor().update_history
            )(history_maker, from_undo, from_redo)
            # fmt: on

            self.main_window.statusBar().showMessage(
                'Node name "{0}" has been changed to "{1}".'.format(
                    old_node_name, new_node_name
                )
            )

    def rename_subprocesses(self, node, parent_node_name):
        """Change the name of a node."""

        if (
            getattr(node.process, "context_name", node.process.name).split(
                "."
            )[0]
            == "Pipeline"
        ):
            if (
                len(
                    getattr(
                        node.process, "context_name", node.process.name
                    ).split(".")
                )
                >= 3
            ):
                node.process.context_name = (
                    "Pipeline."
                    + parent_node_name
                    + "."
                    + ".".join(
                        getattr(
                            node.process, "context_name", node.process.name
                        ).split(".")[2:]
                    )
                )
            else:
                node.process.context_name = "Pipeline." + parent_node_name
        else:
            node.process.context_name = parent_node_name

        if isinstance(node, PipelineNode):
            for name, subnode in node.process.nodes.items():
                if name == "":
                    continue
                else:
                    self.rename_subprocesses(subnode, parent_node_name)

    def filter_attributes(self):
        """Display a filter widget."""
        self.pop_up = AttributesFilter(
            self.project,
            self.scan_list,
            self.process,
            self.node_name,
            "attributes",
            self,
            self.main_window,
        )
        self.pop_up.show()
        self.pop_up.attributes_selected.connect(
            self.update_attributes_from_filter
        )

    def update_attributes_from_filter(self, attributes):
        """Update attributes from filter widget"""
        compl = self.process.completion_engine
        atts = compl.get_attribute_values()
        num_set = 0
        for name, value in attributes.items():
            if atts.trait(name):
                if isinstance(getattr(atts, name), list):
                    setattr(atts, name, value)

                else:
                    setattr(atts, name, value[0])

                num_set += 1

        if num_set == 0 and len(attributes) != 0:
            mbox_icon = QMessageBox.Information
            mbox_title = "Unmatching tags / attributes"
            mbox_text = (
                "The selected data tags do not match the expected "
                "attributes set for process parameters completion"
            )
            mbox = QMessageBox(mbox_icon, mbox_title, mbox_text)
            Qt.QTimer.singleShot(2000, mbox.accept)
            mbox.exec()


class FilterWidget(QWidget):
    """Filter widget used on a Input_Filter process.

    The widget displays a browser with the selected files of the database,
    a rapid search and an advanced search to filter these files. Once the
    filtering is done, the filter is saved in the process.

    .. Methods:
        - layout_view: create the layout
        - ok_clicked: set the filter to the process and closes the widget
        - reset_search_bar: reset the search bar of the rapid search
        - search_str: update the files to display in the browser
        - update_tag_to_filter: update the tag to Filter
        - update_tags: update the list of visualized tags
    """

    def __init__(self, project, node_name, node, main_window):
        """Initialization of the Filter Widget.

        :param project: current project in the software
        :param node_name: name of the current node
        :param node: instance of the corresponding Input_Filter node
        :param main_window: parent main window
        """

        super(FilterWidget, self).__init__(None)

        self.setWindowTitle("Filter - " + node_name)
        self.project = project
        self.visible_tags = self.project.session.get_shown_tags()
        self.node = node
        self.process = node.process
        self.main_window = main_window

        scan_list = []
        # The scan list to filter corresponds to the input of the Input Filter
        if self.process.input and self.process.input is not Undefined:
            for scan in self.process.input:
                path, file_name = os.path.split(scan)
                path, second_folder = os.path.split(path)
                first_folder = os.path.basename(path)
                database_file = os.path.join(
                    first_folder, second_folder, file_name
                )
                scan_list.append(database_file)

        self.scan_list = scan_list

        # Graphical components
        self.table_data = TableDataBrowser(
            self.project, self, self.visible_tags, False, False
        )

        # Reducing the list of scans to selection
        all_scans = self.table_data.scans_to_visualize

        self.table_data.scans_to_visualize = self.scan_list
        self.table_data.scans_to_search = self.scan_list
        self.table_data.update_visualized_rows(all_scans)

        # Filter information
        filter_to_apply = node.process.filter

        # Search
        self.rapid_search = RapidSearch(self)
        if filter_to_apply.search_bar:
            self.rapid_search.setText(filter_to_apply.search_bar)
        self.rapid_search.textChanged.connect(partial(self.search_str))

        self.advanced_search = AdvancedSearch(
            self.project,
            self,
            self.scan_list,
            self.visible_tags,
            from_pipeline=True,
        )
        self.advanced_search.show_search()
        self.advanced_search.apply_filter(filter_to_apply)

        # Initialize Qt objects
        self.button_cross = QToolButton()
        self.push_button_tag_filter = QPushButton(TAG_FILENAME)

        self.layout_view()

    def layout_view(self):
        """Create the layout."""

        sources_images_dir = Config().getSourceImageDir()

        self.button_cross.setStyleSheet("background-color:rgb(255, 255, 255);")
        self.button_cross.setIcon(
            QIcon(os.path.join(sources_images_dir, "gray_cross.png"))
        )
        self.button_cross.clicked.connect(self.reset_search_bar)

        search_bar_layout = QHBoxLayout()
        search_bar_layout.addWidget(self.rapid_search)
        search_bar_layout.addWidget(self.button_cross)

        push_button_tags = QPushButton("Visualized tags")
        push_button_tags.clicked.connect(self.update_tags)

        self.push_button_tag_filter.clicked.connect(self.update_tag_to_filter)

        push_button_ok = QPushButton("OK")
        push_button_ok.clicked.connect(self.ok_clicked)

        push_button_cancel = QPushButton("Cancel")
        push_button_cancel.clicked.connect(self.close)

        # Layout
        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(push_button_tags)
        buttons_layout.addWidget(self.push_button_tag_filter)
        buttons_layout.addStretch(1)
        buttons_layout.addWidget(push_button_ok)
        buttons_layout.addWidget(push_button_cancel)

        main_layout = QVBoxLayout()
        main_layout.addLayout(search_bar_layout)
        main_layout.addWidget(self.advanced_search)
        main_layout.addWidget(self.table_data)
        main_layout.addLayout(buttons_layout)

        self.setLayout(main_layout)

        screen_resolution = QApplication.instance().desktop().screenGeometry()
        width, height = screen_resolution.width(), screen_resolution.height()
        self.setMinimumWidth(round(0.6 * width))
        self.setMinimumHeight(round(0.8 * height))

    def ok_clicked(self):
        """Set the filter to the process and closes the widget."""
        if isinstance(self.process, ProcessMIA):
            (
                fields,
                conditions,
                values,
                links,
                nots,
            ) = self.advanced_search.get_filters(False)
            filt = Filter(
                None,
                nots,
                values,
                fields,
                links,
                conditions,
                self.rapid_search.text(),
            )
            self.process.filter = filt

        # self.set_output_value()
        self.close()

    def reset_search_bar(self):
        """Reset the search bar of the rapid search."""
        self.rapid_search.setText("")
        self.advanced_search.rows = []
        self.advanced_search.show_search()

        # All rows reput
        old_scan_list = self.table_data.scans_to_visualize
        self.table_data.scans_to_visualize = self.scan_list
        self.table_data.scans_to_search = self.scan_list
        self.table_data.update_visualized_rows(old_scan_list)

    def search_str(self, str_search):
        """Update the files to display in the browser.

        :param str_search: string typed in the rapid search
        """

        old_scan_list = self.table_data.scans_to_visualize

        # Every scan taken if empty search
        if str_search == "":
            return_list = self.table_data.scans_to_search
        else:
            # Scans with at least a not defined value
            if str_search == not_defined_value:
                filter = self.prepare_not_defined_filter(
                    self.project.session.get_shown_tags()
                )
            # Scans matching the search
            else:
                filter = self.rapid_search.prepare_filter(
                    str_search,
                    self.project.session.get_shown_tags(),
                    old_scan_list,
                )

            generator = self.project.session.filter_documents(
                COLLECTION_CURRENT, filter
            )

            # Creating the list of scans
            return_list = [getattr(scan, TAG_FILENAME) for scan in generator]

        self.table_data.scans_to_visualize = return_list
        self.advanced_search.scans_list = return_list

        # Rows updated
        self.table_data.update_visualized_rows(old_scan_list)

    #    def set_output_value(self):
    #        """Set the output of the filter to the output of the node."""
    #
    #        result_names = []
    #        filter = self.table_data.get_current_filter()
    #        for i in range(len(filter)):
    #            scan_name = filter[i]
    #            tag_name = self.push_button_tag_filter.text().replace('&', '')
    #            value = self.project.session.get_value(COLLECTION_CURRENT,
    #                                                   scan_name, tag_name)
    #            if tag_name == TAG_FILENAME:
    #                value = os.path.abspath(os.path.join(self.project.folder,
    #                                                     value))
    #            result_names.append(value)
    #
    #        result_files = []
    #        for result_name in result_names:
    #            full_path = os.path.abspath(os.path.join(self.project.folder,
    #                                                     result_name))
    #            result_files.append(full_path)
    #
    #        self.node.set_plug_value("output", result_files)

    def update_tag_to_filter(self):
        """Update the tag to Filter."""

        pop_up = PopUpSelectTagCountTable(
            self.project, self.visible_tags, self.push_button_tag_filter.text()
        )
        if pop_up.exec_():
            self.push_button_tag_filter.setText(pop_up.selected_tag)

    def update_tags(self):
        """Update the list of visualized tags."""

        dialog = QDialog()
        visualized_tags = PopUpVisualizedTags(self.project, self.visible_tags)
        layout = QVBoxLayout()
        layout.addWidget(visualized_tags)
        buttons_layout = QHBoxLayout()
        buttons = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        buttons.accepted.connect(dialog.accept)
        buttons.rejected.connect(dialog.reject)
        buttons_layout.addWidget(buttons)
        layout.addLayout(buttons_layout)
        dialog.setLayout(layout)
        dialog.show()
        dialog.setMinimumHeight(600)
        dialog.setMinimumWidth(600)
        if dialog.exec():
            new_visibilities = []
            for x in range(visualized_tags.list_widget_selected_tags.count()):
                visible_tag = visualized_tags.list_widget_selected_tags.item(
                    x
                ).text()
                new_visibilities.append(visible_tag)
            new_visibilities.append(TAG_FILENAME)
            self.table_data.update_visualized_columns(
                self.visible_tags, new_visibilities
            )
            # self.node_controller.visibles_tags = new_visibilities
            for row in self.advanced_search.rows:
                fields = row[2]
                fields.clear()
                for visible_tag in new_visibilities:
                    fields.addItem(visible_tag)
                fields.model().sort(0)
                fields.addItem("All visualized tags")


# Node controller V1 style
class NodeController(QWidget):
    """
    Allow to change the input and output values of a pipeline node

    .. Methods:
        - clearLayout: clear the layouts of the widget
        - display_filter: display a filter widget
        - display_parameters: display the parameters of the selected node
        - get_index_from_plug_name: return the index of the plug label.
        - update_node_name: update the name of the selected node
        - rename_subprocesses: change the name of a node
        - update_parameters: update the parameters values
        - update_plug_value: update the value of a node plug
        - update_plug_value_from_filter: update the plug value from a filter
           result
        - release_process: remove notification from process (not implemented)

    """

    value_changed = pyqtSignal(list)

    def __init__(self, project, scan_list, pipeline_manager_tab, main_window):
        """
        Initialization of the Node Controller

        :param project: current project in the software
        :param scan_list: list of the selected database files
        :param pipeline_manager_tab: parent widget
        :param main_window: main window of the software
        """

        super(NodeController, self).__init__(pipeline_manager_tab)
        self.project = project
        self.scan_list = scan_list
        self.main_window = main_window
        self.node_name = ""
        self.pipeline = (
            pipeline_manager_tab.pipelineEditorTabs.get_current_pipeline()
        )

        # Layouts
        self.v_box_final = QVBoxLayout()
        self.h_box_node_name = QHBoxLayout()

    def clearLayout(self, layout):
        """Clear the layouts of the widget.

        :param layout: widget with a layout
        """

        for i in reversed(range(len(layout.children()))):
            if type(layout.layout().itemAt(i)) == QtWidgets.QWidgetItem:
                layout.layout().itemAt(i).widget().setParent(None)
            if (
                type(layout.layout().itemAt(i)) == QtWidgets.QHBoxLayout
                or type(layout.layout().itemAt(i)) == QtWidgets.QVBoxLayout
            ):
                layout.layout().itemAt(i).deleteLater()
                for j in reversed(range(len(layout.layout().itemAt(i)))):
                    layout.layout().itemAt(i).itemAt(j).widget().setParent(
                        None
                    )

        if layout.layout() is not None:
            sip.delete(layout.layout())

    def display_filter(self, node_name, plug_name, parameters, process):
        """Display a filter widget.

        :param node_name: name of the node
        :param plug_name: name of the plug
        :param parameters: tuple containing the index of the plug, the current
                           pipeline instance and the type of the plug value
        :param process: process of the node
        """
        self.pop_up = PlugFilter(
            self.project,
            self.scan_list,
            process,
            node_name,
            plug_name,
            self,
            self.main_window,
        )
        self.pop_up.show()
        self.pop_up.plug_value_changed.connect(
            partial(self.update_plug_value_from_filter, plug_name, parameters)
        )

    # def display_parameters(self, node_name: object,
    #                        process: object, pipeline: object) -> object:
    def display_parameters(self, node_name, process, pipeline):
        """Display the parameters of the selected node.

        The node parameters are read and line labels/line edits/push buttons
        are created for each of them. This methods consists mainly in widget
        and layout organization.

        :param node_name: name of the node
        :param process: process of the node
        :param pipeline: current pipeline
        """
        self.node_name = node_name
        self.current_process = process

        self.line_edit_input = []
        self.line_edit_output = []
        self.labels_input = []
        self.labels_output = []

        # Refreshing the layouts
        if len(self.children()) > 0:
            self.clearLayout(self)

        self.v_box_final = QVBoxLayout()

        # Node name
        label_node_name = QLabel()
        label_node_name.setText("Node name:")

        self.line_edit_node_name = QLineEdit()
        self.pipeline = pipeline
        # The pipeline global inputs and outputs node name cannot be modified
        if self.node_name not in ("inputs", "outputs"):
            self.line_edit_node_name.setText(self.node_name)
            self.line_edit_node_name.returnPressed.connect(
                self.update_node_name
            )

        else:
            self.line_edit_node_name.setText("Pipeline inputs/outputs")
            self.line_edit_node_name.setReadOnly(True)

        self.h_box_node_name = QHBoxLayout()
        self.h_box_node_name.addWidget(label_node_name)
        self.h_box_node_name.addWidget(self.line_edit_node_name)

        # Inputs
        self.button_group_inputs = QGroupBox("Inputs")
        self.v_box_inputs = QVBoxLayout()
        idx = 0

        for name, trait in process.user_traits().items():
            if name == "nodes_activation":
                continue

            if trait.userlevel is not None and trait.userlevel > 0:
                continue

            if not trait.output:
                label_input = QLabel()
                label_input.setText(str(name))
                self.labels_input.insert(idx, label_input)
                try:
                    value = getattr(process, name)
                except TraitError:
                    value = Undefined
                trait_type = trait_ids(process.trait(name))

                self.line_edit_input.insert(idx, QLineEdit())
                self.line_edit_input[idx].setText(str(value))
                self.line_edit_input[idx].returnPressed.connect(
                    partial(
                        self.update_plug_value,
                        "in",
                        name,
                        pipeline,
                        type(value),
                    )
                )

                h_box = QHBoxLayout()
                h_box.addWidget(label_input)
                h_box.addWidget(self.line_edit_input[idx])

                # Adding the possibility to filter pipeline global
                # inputs except if the input is "database_scans"
                # which means that the scans will be filtered with InputFilter
                if self.node_name == "inputs" and name != "database_scans":
                    if (
                        "File" in trait_type
                        or "List_File" in trait_type
                        or "Any" in trait_type
                    ):
                        parameters = (idx, pipeline, type(value))
                        push_button = QPushButton("Filter")
                        push_button.clicked.connect(
                            partial(
                                self.display_filter,
                                self.node_name,
                                name,
                                parameters,
                                process,
                            )
                        )
                        h_box.addWidget(push_button)

                self.v_box_inputs.addLayout(h_box)

                idx += 1

        self.button_group_inputs.setLayout(self.v_box_inputs)

        # Outputs
        self.button_group_outputs = QGroupBox("Outputs")
        self.v_box_outputs = QVBoxLayout()
        idx = 0
        for name, trait in process.traits(output=True).items():
            if trait.userlevel is not None and trait.userlevel > 0:
                continue

            label_output = QLabel()
            label_output.setText(str(name))
            self.labels_output.insert(idx, label_output)

            value = getattr(process, name)
            trait_type = trait_ids(process.trait(name))

            self.line_edit_output.insert(idx, QLineEdit())
            self.line_edit_output[idx].setText(str(value))
            self.line_edit_output[idx].returnPressed.connect(
                partial(
                    self.update_plug_value, "out", name, pipeline, type(value)
                )
            )

            h_box = QHBoxLayout()
            h_box.addWidget(label_output)
            h_box.addWidget(self.line_edit_output[idx])

            self.v_box_outputs.addLayout(h_box)

            idx += 1

        self.button_group_outputs.setLayout(self.v_box_outputs)

        self.v_box_final.addLayout(self.h_box_node_name)
        self.v_box_final.addWidget(self.button_group_inputs)
        self.v_box_final.addWidget(self.button_group_outputs)
        # fmt: off
        (
            self.main_window.pipeline_manager.pipelineEditorTabs.
            get_current_editor
        )().node_parameters_tmp[node_name] = {}

        (
            self.main_window.pipeline_manager.pipelineEditorTabs.
            get_current_editor
        )().node_parameters_tmp[node_name]["inputs"] = [
            x.text() for x in self.line_edit_input
        ]

        (
            self.main_window.pipeline_manager.pipelineEditorTabs.
            get_current_editor
        )().node_parameters_tmp[node_name]["outputs"] = [
            x.text() for x in self.line_edit_output
        ]

        if (
            "outputs" in
                self.main_window.pipeline_manager.pipelineEditorTabs.
                get_current_editor().node_parameters_tmp
        ):
            del (
                self.main_window.pipeline_manager.pipelineEditorTabs.
                get_current_editor().node_parameters_tmp["outputs"]
            )
        # fmt: on

        # Commented on January, 4th 2020
        # if self.main_window.pipeline_manager.pipelineEditorTabs\
        # .get_current_editor().node_parameters_tmp ==  \
        #         self.main_window.pipeline_manager.pipelineEditorTabs\
        # .get_current_editor().node_parameters and \
        #         self.main_window.pipeline_manager.test_init and \
        #         self.main_window.pipeline_manager.pipelineEditorTabs\
        #                 .get_current_editor().initialized:
        #     self.main_window.pipeline_manager.run_pipeline_action \
        #         .setDisabled(False)
        # else:
        #    self.main_window.pipeline_manager.run_pipeline_action\
        #         .setDisabled(True)
        # if self.main_window.pipeline_manager.iterationTable\
        #         .check_box_iterate.isChecked():
        #     self.main_window.pipeline_manager.run_pipeline_action\
        #         .setDisabled(False)

        self.main_window.pipeline_manager.run_pipeline_action.setDisabled(
            False
        )
        # End - commented on January, 4th 2020

        self.setLayout(self.v_box_final)

    def get_index_from_plug_name(self, plug_name, in_or_out):
        """Return the index of the plug label.

        :param plug_name: name of the plug
        :param in_or_out: "in" if the plug is an input plug, "out" else
        :return: the corresponding index
        """

        if in_or_out == "in":
            for idx, label in enumerate(self.labels_input):
                if label.text() == plug_name:
                    return idx

        else:
            for idx, label in enumerate(self.labels_output):
                if label.text() == plug_name:
                    return idx

    def update_node_name(self, new_node_name=None):
        """Change the name of the selected node and updates the pipeline.

        Because the nodes are stored in a dictionary, we have to create
        a new node that has the same traits as the selected one and create
        new links that are the same than the selected node.

        :param new_node_name: new node name (is None except when this method
                              is called from an undo/redo)
        """
        # Copying the old node
        old_node_name = self.node_name

        if not new_node_name:
            new_node_name = self.line_edit_node_name.text()

        if isinstance(
            self.pipeline.list_process_in_pipeline[0], ProcessIteration
        ):
            if not new_node_name.startswith("iterated_"):
                new_node_name = "iterated_" + new_node_name
                self.line_edit_node_name.setText(new_node_name)

        if new_node_name in list(self.pipeline.nodes.keys()):
            print("Node name already in pipeline")

        else:
            self.pipeline.rename_node(old_node_name, new_node_name)
            self.rename_subprocesses(
                self.pipeline.nodes[new_node_name], new_node_name
            )

            # Updating the node_name attribute
            self.node_name = new_node_name

            # To undo/redo
            self.value_changed.emit(
                [
                    "node_name",
                    self.pipeline.nodes[new_node_name],
                    new_node_name,
                    old_node_name,
                ]
            )

            self.main_window.statusBar().showMessage(
                'Node name "{0}" has been changed to "{1}".'.format(
                    old_node_name, new_node_name
                )
            )

    def rename_subprocesses(self, node, parent_node_name):
        """Change the name of a node."""

        if (
            getattr(node.process, "context_name", node.process.name).split(
                "."
            )[0]
            == "Pipeline"
        ):
            if (
                len(
                    getattr(
                        node.process, "context_name", node.process.name
                    ).split(".")
                )
                >= 3
            ):
                node.process.context_name = (
                    "Pipeline."
                    + parent_node_name
                    + "."
                    + ".".join(
                        getattr(
                            node.process, "context_name", node.process.name
                        ).split(".")[2:]
                    )
                )
            else:
                node.process.context_name = "Pipeline." + parent_node_name
        else:
            node.process.context_name = parent_node_name

        if isinstance(node, PipelineNode):
            for name, subnode in node.process.nodes.items():
                if name == "":
                    continue
                else:
                    self.rename_subprocesses(subnode, parent_node_name)

    def update_parameters(self, process=None):
        """Update the parameters values.

        :param process: process of the node
        """

        # commented on January, 4th 2020
        # self.main_window.pipeline_manager.run_pipeline_action.setDisabled(
        #                                                                True)

        if process is None:
            try:
                process = self.current_process
            except AttributeError:
                # if no node has been clicked, no need to update the widget
                return

        idx = 0
        for name, trait in process.user_traits().items():
            if name == "nodes_activation":
                continue
            if not trait.output:
                try:
                    value = getattr(process, name)
                except TraitError:
                    value = Undefined
                if idx < len(self.line_edit_input):
                    self.line_edit_input[idx].setText(str(value))
                    idx += 1

        idx = 0

        for name, trait in process.traits(output=True).items():
            value = getattr(process, name)
            if idx < len(self.line_edit_output):
                self.line_edit_output[idx].setText(str(value))
                idx += 1

        # self.main_window.pipeline_manager.pipelineEditorTabs\
        #     .update_current_node()

    def update_plug_value(
        self, in_or_out, plug_name, pipeline, value_type, new_value=None
    ):
        """Update the value of a node plug.

        :param in_or_out: "in" if the plug is an input plug, "out" else
        :param plug_name: name of the plug
        :param pipeline: current pipeline
        :param value_type: type of the plug value
        :param new_value: new value for the plug (is None except when this
                          method is called from an undo/redo)
        """

        index = self.get_index_from_plug_name(plug_name, in_or_out)

        # Reading the value from the plug's line edit
        if not new_value:
            if in_or_out == "in":
                new_value = self.line_edit_input[index].text()

            elif in_or_out == "out":
                new_value = self.line_edit_output[index].text()

            else:
                new_value = None

            try:
                new_value = eval(new_value)

            # We try to handle the undefined value with the eval() function
            # See FixME below.
            except SyntaxError:
                new_value = new_value.replace("<undefined>", "'<undefined>'")

                try:
                    new_value = eval(new_value)

                except Exception as err:
                    print("{0}: {1}.".format(err.__class__, err))

            except NameError:
                pass

            except Exception as err:
                print("{0}: {1}.".format(err.__class__, err))

            # except NameError:
            #    print("NameError for value {0}".format(new_value))
            # except SyntaxError:
            #    pass
            #    print("SynthaxError for value {0}".format(new_value))

            if value_type not in [float, int, str, list]:
                value_type = str

        if self.node_name in ["inputs", "outputs"]:
            node_name = ""

        else:
            node_name = self.node_name

        old_value = pipeline.nodes[node_name].get_plug_value(plug_name)

        try:
            # FIXME:Since we replace, above, "<undefined>" with "<undefined>"
            # in order to handle syntax error with eval() we should handle
            # all cases here (big job).
            # For the moment we manage only the dictionary
            if isinstance(new_value, dict):
                new_value = {
                    k: Undefined if v == "<undefined>" else v
                    for k, v in new_value.items()
                }
            pipeline.nodes[node_name].set_plug_value(plug_name, new_value)

        except (TraitError, OSError) as err:
            msg = QMessageBox()
            msg.setText("{}".format(err))
            msg.setIcon(QMessageBox.Warning)
            msg.setWindowTitle(err.__class__.__name__)
            msg.exec_()

            if in_or_out == "in":
                self.line_edit_input[index].setText(str(old_value))
            elif in_or_out == "out":
                self.line_edit_output[index].setText(str(old_value))
            return

        # Update pipeline to "propagate" the node value
        pipeline.update_nodes_and_plugs_activation()

        if in_or_out == "in":
            self.line_edit_input[index].setText(str(new_value))
        elif in_or_out == "out":
            self.line_edit_output[index].setText(str(new_value))

        # To undo/redo
        self.value_changed.emit(
            [
                "plug_value",
                self.node_name,
                old_value,
                plug_name,
                value_type,
                new_value,
            ]
        )

        self.main_window.statusBar().showMessage(
            'Plug "{0}" of node "{1}" has been changed to "{2}".'.format(
                plug_name, node_name, new_value
            )
        )

    def update_plug_value_from_filter(
        self, plug_name, parameters, filter_res_list
    ):
        """Update the plug value from a filter result.

        :param plug_name: name of the plug
        :param parameters: tuple containing the index of the plug, the current
                           pipeline instance and the type of the plug value
        :param filter_res_list: list of the filtered files
        """

        pipeline = parameters[1]
        value_type = parameters[2]

        # If the list contains only one element, setting
        # this element as the plug value
        len_list = len(filter_res_list)

        if len_list > 1:
            res = filter_res_list

        elif len_list == 1:
            res = filter_res_list[0]

        else:
            res = []

        self.update_plug_value("in", plug_name, pipeline, value_type, res)

    def release_process(self):
        """
        Remove notification from process
        """
        # only implemented in CapsulNodeController
        pass



# -*- coding: utf-8 -*-
"""
Utilities and tools used across Mia

Contains:
    Modules:
        - utils.py
"""

###############################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
###############################################################################
from .utils import (  # noqa: F401
    PackagesInstall,
    check_python_version,
    check_value_type,
    launch_mia,
    message_already_exists,
    set_filters_directory_as_default,
    set_item_data,
    set_projects_directory_as_default,
    table_to_database,
    verCmp,
    verify_processes,
    verify_setup,
)
